""" StorageManagementDB is a front end to the Stager Database.

    There are five tables in the StorageManagementDB: Tasks, CacheReplicas, TaskReplicas, StageRequests.

    The Tasks table is the place holder for the tasks that have requested files to be staged.
    These can be from different systems and have different associated call back methods.
    The CacheReplicas table keeps the information on all the CacheReplicas in the system.
    It maps all the file information LFN, PFN, SE to an assigned ReplicaID.
    The TaskReplicas table maps the TaskIDs from the Tasks table to the ReplicaID from the CacheReplicas table.
    The StageRequests table contains each of the prestage request IDs for each of the replicas.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from past.builtins import long
import six
import inspect
import threading

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities.List import intListToString, stringListToString

# Stage Request are issue with a length of "PinLength"
# However, once Staged, the entry in the StageRequest will set a PinExpiryTime only for "PinLength" / THROTTLING_STEPS
# As PinExpiryTime arrives, StageRequest and their corresponding CacheReplicas entries are cleaned
# This allows to throttle the submission of Stage Requests up to a maximum of "DiskCacheTB" per "PinLength"
# After "PinLength" / THROTTLING_STEPS seconds, entries are removed, so new requests for the same replica will trigger
# a new Stage Request to the SE, and thus an update of the Pinning on the SE.
#
#  - "PinLength" is an Option of the StageRequest Agent that defaults to THROTTLING_TIME
#  - "DiskCacheTB" is an Option of the StorageElement that defaults to 1 (TB)
#
THROTTLING_TIME = 86400
THROTTLING_STEPS = 12


class StorageManagementDB(DB):
    def __init__(self, systemInstance="Default"):
        DB.__init__(self, "StorageManagementDB", "StorageManagement/StorageManagementDB")
        self.lock = threading.Lock()
        self.TASKPARAMS = [
            "TaskID",
            "Status",
            "Source",
            "SubmitTime",
            "LastUpdate",
            "CompleteTime",
            "CallBackMethod",
            "SourceTaskID",
        ]
        self.REPLICAPARAMS = [
            "ReplicaID",
            "Type",
            "Status",
            "SE",
            "LFN",
            "PFN",
            "Size",
            "FileChecksum",
            "GUID",
            "SubmitTime",
            "LastUpdate",
            "Reason",
            "Links",
        ]
        self.STAGEPARAMS = [
            "ReplicaID",
            "StageStatus",
            "RequestID",
            "StageRequestSubmitTime",
            "StageRequestCompletedTime",
            "PinLength",
            "PinExpiryTime",
        ]
        self.STATES = ["Failed", "New", "Waiting", "Offline", "StageSubmitted", "Staged"]

    def __getConnection(self, connection):
        if connection:
            return connection
        res = self._getConnection()
        if res["OK"]:
            return res["Value"]
        gLogger.warn("Failed to get MySQL connection", res["Message"])
        return connection

    def _caller(self):
        return inspect.stack()[2][3]

    ################################################################
    #
    # State machine management
    #

    def updateTaskStatus(self, taskIDs, newTaskStatus, connection=False):
        return self.__updateTaskStatus(taskIDs, newTaskStatus, connection=connection)

    def __updateTaskStatus(self, taskIDs, newTaskStatus, force=False, connection=False):
        connection = self.__getConnection(connection)
        if not taskIDs:
            return S_OK(taskIDs)
        if force:
            toUpdate = taskIDs
        else:
            res = self._checkTaskUpdate(taskIDs, newTaskStatus, connection=connection)
            if not res["OK"]:
                return res
            toUpdate = res["Value"]
        if not toUpdate:
            return S_OK(toUpdate)

        # reqSelect = "SELECT * FROM Tasks WHERE TaskID IN (%s) AND Status != '%s';" % (
        #     intListToString( toUpdate ), newTaskStatus
        # )
        reqSelect = "SELECT TaskID FROM Tasks WHERE TaskID IN (%s) AND Status != '%s';" % (
            intListToString(toUpdate),
            newTaskStatus,
        )
        resSelect = self._query(reqSelect, connection)
        if not resSelect["OK"]:
            gLogger.error(
                "%s.%s_DB: problem retrieving record:" % (self._caller(), "__updateTaskStatus"),
                "%s. %s" % (reqSelect, resSelect["Message"]),
            )

        req = "UPDATE Tasks SET Status='%s',LastUpdate=UTC_TIMESTAMP() WHERE TaskID IN (%s) AND Status != '%s';" % (
            newTaskStatus,
            intListToString(toUpdate),
            newTaskStatus,
        )
        res = self._update(req, connection)
        if not res["OK"]:
            return res

        taskIDs = []
        for record in resSelect["Value"]:
            taskIDs.append(record[0])
            gLogger.verbose("%s.%s_DB: to_update Tasks =  %s" % (self._caller(), "__updateTaskStatus", record))

        if len(taskIDs) > 0:
            reqSelect1 = "SELECT * FROM Tasks WHERE TaskID IN (%s);" % intListToString(taskIDs)
            resSelect1 = self._query(reqSelect1, connection)
            if not resSelect1["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving records: %s. %s"
                    % (self._caller(), "__updateTaskStatus", reqSelect1, resSelect1["Message"])
                )
            else:
                for record in resSelect1["Value"]:
                    gLogger.verbose("%s.%s_DB: updated Tasks = %s" % (self._caller(), "__updateTaskStatus", record))

        return S_OK(toUpdate)

    def _checkTaskUpdate(self, taskIDs, newTaskState, connection=False):
        connection = self.__getConnection(connection)
        if not taskIDs:
            return S_OK(taskIDs)
        # * -> Failed
        if newTaskState == "Failed":
            oldTaskState = []
        # StageCompleting -> Done
        elif newTaskState == "Done":
            oldTaskState = ["StageCompleting"]
        # StageSubmitted -> StageCompleting
        elif newTaskState == "StageCompleting":
            oldTaskState = ["StageSubmitted"]
        # Waiting -> StageSubmitted
        elif newTaskState == "StageSubmitted":
            oldTaskState = ["Waiting", "Offline"]
        # New -> Waiting
        elif newTaskState == "Waiting":
            oldTaskState = ["New"]
        elif newTaskState == "Offline":
            oldTaskState = ["Waiting"]
        else:
            return S_ERROR("Task status not recognized")
        if not oldTaskState:
            toUpdate = taskIDs
        else:
            req = "SELECT TaskID FROM Tasks WHERE Status in (%s) AND TaskID IN (%s)" % (
                stringListToString(oldTaskState),
                intListToString(taskIDs),
            )
            res = self._query(req, connection)
            if not res["OK"]:
                return res
            toUpdate = [row[0] for row in res["Value"]]
        return S_OK(toUpdate)

    def updateReplicaStatus(self, replicaIDs, newReplicaStatus, connection=False):
        connection = self.__getConnection(connection)
        if not replicaIDs:
            return S_OK(replicaIDs)
        res = self._checkReplicaUpdate(replicaIDs, newReplicaStatus)
        if not res["OK"]:
            return res
        toUpdate = res["Value"]
        if not toUpdate:
            return S_OK(toUpdate)
        # reqSelect = "SELECT * FROM CacheReplicas WHERE ReplicaID IN (%s) AND Status != '%s';" % (
        #     intListToString( toUpdate ), newReplicaStatus
        # )
        reqSelect = "SELECT ReplicaID FROM CacheReplicas WHERE ReplicaID IN (%s) AND Status != '%s';" % (
            intListToString(toUpdate),
            newReplicaStatus,
        )
        resSelect = self._query(reqSelect, connection)
        if not resSelect["OK"]:
            gLogger.error(
                "%s.%s_DB: problem retrieving record:" % (self._caller(), "updateReplicaStatus"),
                "%s. %s" % (reqSelect, resSelect["Message"]),
            )

        req = (
            "UPDATE CacheReplicas SET Status='%s',LastUpdate=UTC_TIMESTAMP() WHERE ReplicaID IN (%s) AND Status != '%s';"
            % (newReplicaStatus, intListToString(toUpdate), newReplicaStatus)
        )
        res = self._update(req, connection)
        if not res["OK"]:
            return res

        replicaIDs = []
        for record in resSelect["Value"]:
            replicaIDs.append(record[0])
            gLogger.verbose("%s.%s_DB: to_update CacheReplicas =  %s" % (self._caller(), "updateReplicaStatus", record))
        if len(replicaIDs) > 0:
            reqSelect1 = "SELECT * FROM CacheReplicas WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
            resSelect1 = self._query(reqSelect1, connection)
            if not resSelect1["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving records: %s. %s"
                    % (self._caller(), "updateReplicaStatus", reqSelect1, resSelect1["Message"])
                )
            else:
                for record in resSelect1["Value"]:
                    gLogger.verbose(
                        "%s.%s_DB: updated CacheReplicas = %s" % (self._caller(), "updateReplicaStatus", record)
                    )

        res = self._updateTasksForReplica(replicaIDs, connection=connection)
        if not res["OK"]:
            return res
        return S_OK(toUpdate)

    def _updateTasksForReplica(self, replicaIDs, connection=False):
        tasksInStatus = {}
        for state in self.STATES:
            tasksInStatus[state] = []

        req = (
            "SELECT T.TaskID,T.Status FROM Tasks AS T, TaskReplicas AS R WHERE R.ReplicaID IN "
            "( %s ) AND R.TaskID = T.TaskID GROUP BY T.TaskID, T.Status;"
        ) % intListToString(replicaIDs)
        res = self._query(req, connection)
        if not res["OK"]:
            return res

        for taskId, status in res["Value"]:
            subreq = (
                "SELECT DISTINCT(C.Status) FROM TaskReplicas AS R, CacheReplicas AS C WHERE R.TaskID=%s AND R.ReplicaID = C.ReplicaID;"
                % taskId
            )
            subres = self._query(subreq, connection)
            if not subres["OK"]:
                return subres

            cacheStatesForTask = [row[0] for row in subres["Value"]]
            if not cacheStatesForTask:
                tasksInStatus["Failed"].append(taskId)
                continue

            wrongState = False
            for state in cacheStatesForTask:
                if state not in self.STATES:
                    wrongState = True
                    break
            if wrongState:
                tasksInStatus["Failed"].append(taskId)
                continue
            for state in self.STATES:
                if state in cacheStatesForTask:
                    if status != state:
                        tasksInStatus[state].append(taskId)
                    break

        for newStatus in tasksInStatus:
            if tasksInStatus[newStatus]:
                res = self.__updateTaskStatus(tasksInStatus[newStatus], newStatus, True, connection=connection)
                if not res["OK"]:
                    gLogger.warn("Failed to update task associated to replicas", res["Message"])
                    # return res
        return S_OK(tasksInStatus)

    def getAssociatedReplicas(self, replicaIDs):
        """Retrieve the list of Replicas that belong to the same Tasks as the provided list"""
        res = self._getReplicaIDTasks(replicaIDs)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getAssociatedReplicas: Failed to get Tasks.", res["Message"])
            return res
        taskIDs = res["Value"]

        return self.getCacheReplicas({"TaskID": taskIDs})

    def _checkReplicaUpdate(self, replicaIDs, newReplicaState, connection=False):
        connection = self.__getConnection(connection)
        if not replicaIDs:
            return S_OK(replicaIDs)
        # * -> Failed
        if newReplicaState == "Failed":
            oldReplicaState = []
        # New -> Waiting
        elif newReplicaState == "Waiting":
            oldReplicaState = ["New"]
        # Waiting -> StageSubmitted
        elif newReplicaState == "StageSubmitted":
            oldReplicaState = ["Waiting", "Offline"]
        # StageSubmitted -> Staged
        elif newReplicaState == "Staged":
            oldReplicaState = ["StageSubmitted"]
        elif newReplicaState == "Offline":
            oldReplicaState = ["Waiting"]
        else:
            return S_ERROR("Replica status not recognized")
        if not oldReplicaState:
            toUpdate = replicaIDs
        else:
            req = "SELECT ReplicaID FROM CacheReplicas WHERE Status IN (%s) AND ReplicaID IN (%s)" % (
                stringListToString(oldReplicaState),
                intListToString(replicaIDs),
            )
            res = self._query(req, connection)
            if not res["OK"]:
                return res
            toUpdate = [row[0] for row in res["Value"]]
        return S_OK(toUpdate)

    def __getTaskStateFromReplicaState(self, replicaState):
        # For the moment the task state just references to the replicaState
        return replicaState

    def updateStageRequestStatus(self, replicaIDs, newStageStatus, connection=False):
        connection = self.__getConnection(connection)
        if not replicaIDs:
            return S_OK(replicaIDs)
        res = self._checkStageUpdate(replicaIDs, newStageStatus, connection=connection)
        if not res["OK"]:
            return res
        toUpdate = res["Value"]
        if not toUpdate:
            return S_OK(toUpdate)
        # reqSelect = "Select * FROM CacheReplicas WHERE ReplicaID IN (%s) AND Status != '%s';" % ( intListToString( toUpdate ), newStageStatus )
        reqSelect = "Select ReplicaID FROM CacheReplicas WHERE ReplicaID IN (%s) AND Status != '%s';" % (
            intListToString(toUpdate),
            newStageStatus,
        )
        resSelect = self._query(reqSelect, connection)
        if not resSelect["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "updateStageRequestStatus", reqSelect, resSelect["Message"])
            )

        req = (
            "UPDATE CacheReplicas SET Status='%s',LastUpdate=UTC_TIMESTAMP() WHERE ReplicaID IN (%s) AND Status != '%s';"
            % (newStageStatus, intListToString(toUpdate), newStageStatus)
        )
        res = self._update(req, connection)
        if not res["OK"]:
            return res

        replicaIDs = []
        for record in resSelect["Value"]:
            replicaIDs.append(record[0])
            gLogger.verbose(
                "%s.%s_DB: to_update CacheReplicas =  %s" % (self._caller(), "updateStageRequestStatus", record)
            )

        reqSelect1 = "SELECT * FROM CacheReplicas WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
        resSelect1 = self._query(reqSelect1, connection)
        if not resSelect1["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving records: %s. %s"
                % (self._caller(), "updateStageRequestStatus", reqSelect1, resSelect1["Message"])
            )
        else:
            for record in resSelect1["Value"]:
                gLogger.verbose(
                    "%s.%s_DB: updated CacheReplicas = %s" % (self._caller(), "updateStageRequestStatus", record)
                )

        # Now update the replicas associated to the replicaIDs
        newReplicaStatus = self.__getReplicaStateFromStageState(newStageStatus)
        res = self.updateReplicaStatus(toUpdate, newReplicaStatus, connection=connection)
        if not res["OK"]:
            gLogger.warn("Failed to update cache replicas associated to stage requests", res["Message"])
        return S_OK(toUpdate)

    def _checkStageUpdate(self, replicaIDs, newStageState, connection=False):
        connection = self.__getConnection(connection)
        if not replicaIDs:
            return S_OK(replicaIDs)
        # * -> Failed
        if newStageState == "Failed":
            oldStageState = []
        elif newStageState == "Staged":
            oldStageState = ["StageSubmitted"]
        else:
            return S_ERROR("StageRequest status not recognized")
        if not oldStageState:
            toUpdate = replicaIDs
        else:
            req = "SELECT ReplicaID FROM StageRequests WHERE StageStatus = '%s' AND ReplicaID IN (%s)" % (
                oldStageState,
                intListToString(replicaIDs),
            )
            res = self._query(req, connection)
            if not res["OK"]:
                return res
            toUpdate = [row[0] for row in res["Value"]]
        return S_OK(toUpdate)

    def __getReplicaStateFromStageState(self, stageState):
        # For the moment the replica state just references to the stage state
        return stageState

    #
    #                               End of state machine management
    #
    ################################################################

    ################################################################
    #
    # Monitoring of stage tasks
    #
    def getTaskStatus(self, taskID, connection=False):
        """Obtain the task status from the Tasks table."""
        connection = self.__getConnection(connection)
        res = self.getTaskInfo(taskID, connection=connection)
        if not res["OK"]:
            return res
        taskInfo = res["Value"][taskID]
        return S_OK(taskInfo["Status"])

    def getTaskInfo(self, taskID, connection=False):
        """Obtain all the information from the Tasks table for a supplied task."""
        connection = self.__getConnection(connection)
        req = (
            "SELECT TaskID,Status,Source,SubmitTime,CompleteTime,CallBackMethod,SourceTaskID from Tasks WHERE TaskID IN (%s);"
            % intListToString(taskID)
        )
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getTaskInfo: Failed to get task information.", res["Message"])
            return res
        resDict = {}
        for taskID, status, source, submitTime, completeTime, callBackMethod, sourceTaskID in res["Value"]:
            resDict[sourceTaskID] = {
                "Status": status,
                "Source": source,
                "SubmitTime": submitTime,
                "CompleteTime": completeTime,
                "CallBackMethod": callBackMethod,
                "SourceTaskID": sourceTaskID,
            }
        if not resDict:
            gLogger.error("StorageManagementDB.getTaskInfo: The supplied task did not exist", taskID)
            return S_ERROR("The supplied task %s did not exist" % taskID)
        return S_OK(resDict)

    def _getTaskIDForJob(self, jobID, connection=False):
        # Stager taskID is retrieved from the source DIRAC jobID
        connection = self.__getConnection(connection)
        req = "SELECT TaskID from Tasks WHERE SourceTaskID=%s;" % int(jobID)
        res = self._query(req)
        if not res["OK"]:
            gLogger.error(
                "%s.%s_DB: problem retrieving record:" % (self._caller(), "_getTaskIDForJob"),
                "%s. %s" % (req, res["Message"]),
            )
            return S_ERROR("The supplied JobID does not exist!")
        taskID = [row[0] for row in res["Value"]]
        return S_OK(taskID)

    def getTaskSummary(self, jobID, connection=False):
        """Obtain the task summary from the database."""
        connection = self.__getConnection(connection)
        res = self._getTaskIDForJob(jobID, connection=connection)
        if not res["OK"]:
            return res
        if res["Value"]:
            taskID = res["Value"]
        else:
            return S_OK()
        res = self.getTaskInfo(taskID, connection=connection)
        if not res["OK"]:
            return res
        taskInfo = res["Value"]
        req = (
            "SELECT R.LFN,R.SE,R.PFN,R.Size,R.Status,R.LastUpdate,R.Reason FROM CacheReplicas AS R, TaskReplicas AS TR WHERE TR.TaskID in (%s) AND TR.ReplicaID=R.ReplicaID;"
            % intListToString(taskID)
        )
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getTaskSummary: Failed to get Replica summary for task.", res["Message"])
            return res
        replicaInfo = {}
        for lfn, storageElement, pfn, fileSize, status, lastupdate, reason in res["Value"]:
            replicaInfo[lfn] = {
                "StorageElement": storageElement,
                "PFN": pfn,
                "FileSize": fileSize,
                "Status": status,
                "LastUpdate": lastupdate,
                "Reason": reason,
            }
        resDict = {"TaskInfo": taskInfo, "ReplicaInfo": replicaInfo}
        return S_OK(resDict)

    def getTasks(
        self,
        condDict={},
        older=None,
        newer=None,
        timeStamp="SubmitTime",
        orderAttribute=None,
        limit=None,
        connection=False,
    ):
        """Get stage requests for the supplied selection with support for web standard structure"""
        connection = self.__getConnection(connection)
        req = "SELECT %s FROM Tasks" % (intListToString(self.TASKPARAMS))
        if condDict or older or newer:
            if "ReplicaID" in condDict:
                replicaIDs = condDict.pop("ReplicaID")
                if not isinstance(replicaIDs, (list, tuple)):
                    replicaIDs = [replicaIDs]
                res = self._getReplicaIDTasks(replicaIDs, connection=connection)
                if not res["OK"]:
                    return res
                condDict["TaskID"] = res["Value"]
            req = "%s %s" % (req, self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit))
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        # Cast to list to be able to serialize
        tasks = [list(row) for row in res["Value"]]
        resultDict = {}
        for row in tasks:
            resultDict[row[0]] = dict(zip(self.TASKPARAMS[1:], row[1:]))
        result = S_OK(resultDict)
        result["Records"] = tasks
        result["ParameterNames"] = self.TASKPARAMS
        return result

    def getCacheReplicas(
        self,
        condDict={},
        older=None,
        newer=None,
        timeStamp="LastUpdate",
        orderAttribute=None,
        limit=None,
        connection=False,
    ):
        """Get cache replicas for the supplied selection with support for the web standard structure"""
        connection = self.__getConnection(connection)
        req = "SELECT %s FROM CacheReplicas" % (intListToString(self.REPLICAPARAMS))
        if condDict or older or newer:
            if "TaskID" in condDict:
                taskIDs = condDict.pop("TaskID")
                if not isinstance(taskIDs, (list, tuple)):
                    taskIDs = [taskIDs]
                res = self._getTaskReplicaIDs(taskIDs, connection=connection)
                if not res["OK"]:
                    return res
                if res["Value"]:
                    condDict["ReplicaID"] = res["Value"]
                else:
                    condDict["ReplicaID"] = [-1]
            # BUG: limit is ignored unless there is a nonempty condition dictionary OR
            # older OR newer is nonemtpy
            req = "%s %s" % (req, self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit))
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        # Cast to list to be able to serialize
        cacheReplicas = [list(row) for row in res["Value"]]
        resultDict = {}
        for row in cacheReplicas:
            resultDict[row[0]] = dict(zip(self.REPLICAPARAMS[1:], row[1:]))
        result = S_OK(resultDict)
        result["Records"] = cacheReplicas
        result["ParameterNames"] = self.REPLICAPARAMS
        return result

    def getStageRequests(
        self,
        condDict={},
        older=None,
        newer=None,
        timeStamp="StageRequestSubmitTime",
        orderAttribute=None,
        limit=None,
        connection=False,
    ):
        """Get stage requests for the supplied selection with support for web standard structure"""
        connection = self.__getConnection(connection)
        req = "SELECT %s FROM StageRequests" % (intListToString(self.STAGEPARAMS))
        if condDict or older or newer:
            if "TaskID" in condDict:
                taskIDs = condDict.pop("TaskID")
                if not isinstance(taskIDs, (list, tuple)):
                    taskIDs = [taskIDs]
                res = self._getTaskReplicaIDs(taskIDs, connection=connection)
                if not res["OK"]:
                    return res
                if res["Value"]:
                    condDict["ReplicaID"] = res["Value"]
                else:
                    condDict["ReplicaID"] = [-1]
            req = "%s %s" % (req, self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit))
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        # Cast to list to be able to serialize
        stageRequests = [list(row) for row in res["Value"]]
        resultDict = {}
        for row in stageRequests:
            resultDict[row[0]] = dict(zip(self.STAGEPARAMS[1:], row[1:]))
        result = S_OK(resultDict)
        result["Records"] = stageRequests
        result["ParameterNames"] = self.STAGEPARAMS
        return result

    def _getTaskReplicaIDs(self, taskIDs, connection=False):
        if not taskIDs:
            return S_OK([])
        req = "SELECT DISTINCT(ReplicaID) FROM TaskReplicas WHERE TaskID IN (%s);" % intListToString(taskIDs)
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        replicaIDs = [row[0] for row in res["Value"]]
        return S_OK(replicaIDs)

    def _getReplicaIDTasks(self, replicaIDs, connection=False):
        if not replicaIDs:
            return S_OK([])
        req = "SELECT DISTINCT(TaskID) FROM TaskReplicas WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        taskIDs = [row[0] for row in res["Value"]]

        return S_OK(taskIDs)

    #
    #                               End of monitoring of stage tasks
    #
    ################################################################

    ####################################################################
    #
    # Submission of stage requests
    #

    def setRequest(self, lfnDict, source, callbackMethod, sourceTaskID, connection=False):
        """This method populates the StorageManagementDB Tasks table with the requested files."""
        connection = self.__getConnection(connection)
        if not lfnDict:
            return S_ERROR("No files supplied in request")
        # The first step is to create the task in the Tasks table
        res = self._createTask(source, callbackMethod, sourceTaskID, connection=connection)
        if not res["OK"]:
            return res
        taskID = res["Value"]
        # Get the Replicas which already exist in the CacheReplicas table
        allReplicaIDs = []
        taskStates = []
        for se, lfns in lfnDict.items():
            if isinstance(lfns, six.string_types):
                lfns = [lfns]
            res = self._getExistingReplicas(se, lfns, connection=connection)
            if not res["OK"]:
                return res
            existingReplicas = res["Value"]
            # Insert the CacheReplicas that do not already exist
            for lfn in lfns:
                if lfn in existingReplicas:
                    gLogger.verbose(
                        "StorageManagementDB.setRequest: Replica already exists in CacheReplicas table %s @ %s"
                        % (lfn, se)
                    )
                    existingFileState = existingReplicas[lfn][1]
                    taskState = self.__getTaskStateFromReplicaState(existingFileState)
                else:
                    res = self._insertReplicaInformation(lfn, se, "Stage", connection=connection)
                    if not res["OK"]:
                        self._cleanTask(taskID, connection=connection)
                        return res

                    existingReplicas[lfn] = (res["Value"], "New")
                    newFileState = existingReplicas[lfn][1]
                    taskState = self.__getTaskStateFromReplicaState(newFileState)
                if taskState not in taskStates:
                    taskStates.append(taskState)

            allReplicaIDs.extend(existingReplicas.values())
        # Insert all the replicas into the TaskReplicas table
        res = self._insertTaskReplicaInformation(taskID, allReplicaIDs, connection=connection)
        if not res["OK"]:
            self._cleanTask(taskID, connection=connection)
            return res
        # Check whether the the task status is Done based on the existing file states
        # If all the files for a particular Task are 'Staged', update the Task
        if taskStates == ["Staged"]:
            # so if the tasks are for LFNs from the lfns dictionary, which are already staged,
            # they immediately change state New->Done. Fixed it to translate such tasks to 'Staged' state
            self.__updateTaskStatus([taskID], "Staged", True, connection=connection)
        if "Failed" in taskStates:
            self.__updateTaskStatus([taskID], "Failed", True, connection=connection)
        return S_OK(taskID)

    def _cleanTask(self, taskID, connection=False):
        """Remove a task and any related information"""
        connection = self.__getConnection(connection)
        self.removeTasks([taskID], connection=connection)
        self.removeUnlinkedReplicas(connection=connection)

    def _createTask(self, source, callbackMethod, sourceTaskID, connection=False):
        """Enter the task details into the Tasks table"""
        connection = self.__getConnection(connection)
        req = (
            "INSERT INTO Tasks (Source,SubmitTime,CallBackMethod,SourceTaskID) VALUES ('%s',UTC_TIMESTAMP(),'%s','%s');"
            % (source, callbackMethod, sourceTaskID)
        )
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB._createTask: Failed to create task.", res["Message"])
            return res
        # gLogger.info( "%s_DB:%s" % ('_createTask',req))
        taskID = res["lastRowId"]
        reqSelect = "SELECT * FROM Tasks WHERE TaskID = %s;" % (taskID)
        resSelect = self._query(reqSelect, connection)
        if not resSelect["OK"]:
            gLogger.info(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "_createTask", reqSelect, resSelect["Message"])
            )
        else:
            gLogger.verbose("%s.%s_DB: inserted Tasks = %s" % (self._caller(), "_createTask", resSelect["Value"][0]))

        # gLogger.info("StorageManagementDB._createTask: Created task with ('%s','%s','%s') and obtained TaskID %s" % (source,callbackMethod,sourceTaskID,taskID))
        return S_OK(taskID)

    def _getExistingReplicas(self, storageElement, lfns, connection=False):
        """Obtains the ReplicasIDs for the replicas already entered in the CacheReplicas table"""
        connection = self.__getConnection(connection)
        req = "SELECT ReplicaID,LFN,Status FROM CacheReplicas WHERE SE = '%s' AND LFN IN (%s);" % (
            storageElement,
            stringListToString(lfns),
        )
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB._getExistingReplicas: Failed to get existing replicas.", res["Message"])
            return res
        existingReplicas = {}
        for replicaID, lfn, status in res["Value"]:
            existingReplicas[lfn] = (replicaID, status)
        return S_OK(existingReplicas)

    def _insertReplicaInformation(self, lfn, storageElement, rType, connection=False):
        """Enter the replica into the CacheReplicas table"""
        connection = self.__getConnection(connection)
        req = (
            "INSERT INTO CacheReplicas (Type,SE,LFN,PFN,Size,FileChecksum,GUID,SubmitTime,LastUpdate) VALUES ('%s','%s','%s','',0,'','',UTC_TIMESTAMP(),UTC_TIMESTAMP());"
            % (rType, storageElement, lfn)
        )
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("_insertReplicaInformation: Failed to insert to CacheReplicas table.", res["Message"])
            return res
        # gLogger.info( "%s_DB:%s" % ('_insertReplicaInformation',req))

        replicaID = res["lastRowId"]
        reqSelect = "SELECT * FROM CacheReplicas WHERE ReplicaID = %s;" % (replicaID)
        resSelect = self._query(reqSelect, connection)
        if not resSelect["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "_insertReplicaInformation", reqSelect, resSelect["Message"])
            )
        else:
            gLogger.verbose(
                "%s.%s_DB: inserted CacheReplicas = %s"
                % (self._caller(), "_insertReplicaInformation", resSelect["Value"][0])
            )
        # gLogger.verbose("_insertReplicaInformation: Inserted Replica ('%s','%s') and obtained ReplicaID %s" % (lfn,storageElement,replicaID))
        return S_OK(replicaID)

    def _insertTaskReplicaInformation(self, taskID, replicaIDs, connection=False):
        """Enter the replicas into TaskReplicas table"""
        connection = self.__getConnection(connection)
        req = "INSERT INTO TaskReplicas (TaskID,ReplicaID) VALUES "
        for replicaID, _status in replicaIDs:
            replicaString = "(%s,%s)," % (taskID, replicaID)
            req = "%s %s" % (req, replicaString)
        req = req.rstrip(",")
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error(
                "StorageManagementDB._insertTaskReplicaInformation: Failed to insert to TaskReplicas table.",
                res["Message"],
            )
            return res
        # gLogger.info( "%s_DB:%s" % ('_insertTaskReplicaInformation',req))
        gLogger.verbose(
            "StorageManagementDB._insertTaskReplicaInformation: Successfully added %s CacheReplicas to Task %s."
            % (res["Value"], taskID)
        )
        return S_OK()

    #
    #                               End of insertion methods
    #
    ################################################################

    ####################################################################

    def getStagedReplicas(self, connection=False):
        connection = self.__getConnection(connection)
        req = "SELECT TR.TaskID, R.Status, COUNT(*) from TaskReplicas as TR, CacheReplicas as R where TR.ReplicaID=R.ReplicaID GROUP BY TR.TaskID,R.Status;"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getStagedReplicas: Failed to get eligible TaskReplicas", res["Message"])
            return res
        goodTasks = []
        for taskID, status, _count in res["Value"]:
            if taskID in goodTasks:
                continue
            elif status in ("Staged", "StageSubmitted"):
                goodTasks.append(taskID)
        return self.getCacheReplicas({"Status": "Staged", "TaskID": goodTasks}, connection=connection)

    def getWaitingReplicas(self, connection=False):
        connection = self.__getConnection(connection)
        req = "SELECT TR.TaskID, R.Status, COUNT(*) from TaskReplicas as TR, CacheReplicas as R where TR.ReplicaID=R.ReplicaID GROUP BY TR.TaskID,R.Status;"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getWaitingReplicas: Failed to get eligible TaskReplicas", res["Message"])
            return res
        badTasks = []
        goodTasks = []
        for taskID, status, _count in res["Value"]:
            if taskID in badTasks:
                continue
            elif status in ("New", "Failed"):
                badTasks.append(taskID)
            elif status == "Waiting":
                goodTasks.append(taskID)
        return self.getCacheReplicas({"Status": "Waiting", "TaskID": goodTasks}, connection=connection)

    ####################################################################

    def getOfflineReplicas(self, connection=False):
        connection = self.__getConnection(connection)
        req = "SELECT TR.TaskID, R.Status, COUNT(*) from TaskReplicas as TR, CacheReplicas as R where TR.ReplicaID=R.ReplicaID GROUP BY TR.TaskID,R.Status;"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getOfflineReplicas: Failed to get eligible TaskReplicas", res["Message"])
            return res
        badTasks = []
        goodTasks = []
        for taskID, status, _count in res["Value"]:
            if taskID in badTasks:
                continue
            elif status in ("New", "Failed"):
                badTasks.append(taskID)
            elif status == "Offline":
                goodTasks.append(taskID)
        return self.getCacheReplicas({"Status": "Offline", "TaskID": goodTasks}, connection=connection)

    ####################################################################

    def getTasksWithStatus(self, status):
        """This method retrieves the TaskID from the Tasks table with the supplied Status."""
        req = "SELECT TaskID,Source,CallBackMethod,SourceTaskID from Tasks WHERE Status = '%s';" % status
        res = self._query(req)
        if not res["OK"]:
            return res
        taskIDs = {}
        for taskID, source, callback, sourceTask in res["Value"]:
            taskIDs[taskID] = (source, callback, sourceTask)
        return S_OK(taskIDs)

    ####################################################################
    #
    # The state transition of the CacheReplicas from *->Failed
    #

    def updateReplicaFailure(self, terminalReplicaIDs):
        """This method sets the status to Failure with the failure reason for the supplied Replicas."""
        res = self.updateReplicaStatus(list(terminalReplicaIDs), "Failed")
        if not res["OK"]:
            return res
        updated = res["Value"]
        if not updated:
            return S_OK(updated)
        for replicaID in updated:
            reqSelect = "Select * FROM CacheReplicas WHERE ReplicaID = %d" % (replicaID)
            resSelect = self._query(reqSelect)
            if not resSelect["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving record: %s. %s"
                    % (self._caller(), "updateReplicaFailure", reqSelect, resSelect["Message"])
                )

            req = "UPDATE CacheReplicas SET Reason = '%s' WHERE ReplicaID = %d" % (
                terminalReplicaIDs[replicaID],
                replicaID,
            )
            res = self._update(req)
            if not res["OK"]:
                gLogger.error(
                    "StorageManagementDB.updateReplicaFailure: Failed to update replica fail reason.", res["Message"]
                )
                return res

            replicaIDs = []
            for record in resSelect["Value"]:
                replicaIDs.append(record[0])
                gLogger.verbose(
                    "%s.%s_DB: to_update CacheReplicas =  %s" % (self._caller(), "updateReplicaFailure", record)
                )

            reqSelect1 = "SELECT * FROM CacheReplicas WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
            resSelect1 = self._query(reqSelect1)
            if not resSelect1["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving records: %s. %s"
                    % (self._caller(), "updateReplicaFailure", reqSelect1, resSelect1["Message"])
                )
            else:
                for record in resSelect1["Value"]:
                    gLogger.verbose(
                        "%s.%s_DB: updated CacheReplicas = %s" % (self._caller(), "updateReplicaFailure", record)
                    )

        return S_OK(updated)

    ####################################################################
    #
    # The state transition of the CacheReplicas from New->Waiting
    #

    def updateReplicaInformation(self, replicaTuples):
        """This method set the replica size information and pfn for the requested storage element."""
        for replicaID, pfn, size in replicaTuples:
            # reqSelect = "SELECT * FROM CacheReplicas WHERE ReplicaID = %s and Status != 'Cancelled';" % ( replicaID )
            reqSelect = "SELECT ReplicaID FROM CacheReplicas WHERE ReplicaID = %s and Status != 'Cancelled';" % (
                replicaID
            )
            resSelect = self._query(reqSelect)
            if not resSelect["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving record: %s. %s"
                    % (self._caller(), "updateReplicaInformation", reqSelect, resSelect["Message"])
                )

            req = (
                "UPDATE CacheReplicas SET PFN = '%s', Size = %s, Status = 'Waiting' WHERE ReplicaID = %s and Status != 'Cancelled';"
                % (pfn, size, replicaID)
            )
            res = self._update(req)
            if not res["OK"]:
                gLogger.error(
                    "StagerDB.updateReplicaInformation: Failed to insert replica information.", res["Message"]
                )

            replicaIDs = []
            for record in resSelect["Value"]:
                replicaIDs.append(record[0])
                gLogger.verbose(
                    "%s.%s_DB: to_update CacheReplicas =  %s" % (self._caller(), "updateReplicaInformation", record)
                )

            reqSelect1 = "SELECT * FROM CacheReplicas WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
            resSelect1 = self._query(reqSelect1)
            if not resSelect1["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving record: %s. %s"
                    % (self._caller(), "updateReplicaInformation", reqSelect1, resSelect1["Message"])
                )
            else:
                for record in resSelect1["Value"]:
                    gLogger.verbose(
                        "%s.%s_DB: updated CacheReplicas = %s" % (self._caller(), "updateReplicaInformation", record)
                    )

            gLogger.debug(
                "StagerDB.updateReplicaInformation: Successfully updated CacheReplicas record With Status=Waiting, for ReplicaID= %s"
                % (replicaID)
            )
        return S_OK()

    ####################################################################
    #
    # The state transition of the CacheReplicas from Waiting->StageSubmitted
    #

    def getSubmittedStagePins(self):
        # change the query to take into account pin expiry time
        req = "SELECT SE,COUNT(*),SUM(Size) from CacheReplicas WHERE Status NOT IN ('New','Waiting','Offline','Failed') GROUP BY SE;"
        # req = "SELECT SE,Count(*),SUM(Size) from CacheReplicas,StageRequests WHERE Status NOT IN ('New','Waiting','Failed') and CacheReplicas.ReplicaID=StageRequests.ReplicaID and PinExpiryTime>UTC_TIMESTAMP() GROUP BY SE;"
        res = self._query(req)
        if not res["OK"]:
            gLogger.error(
                "StorageManagementDB.getSubmittedStagePins: Failed to obtain submitted requests.", res["Message"]
            )
            return res
        storageRequests = {}
        for storageElement, replicas, totalSize in res["Value"]:
            storageRequests[storageElement] = {"Replicas": int(replicas), "TotalSize": int(totalSize)}
        return S_OK(storageRequests)

    def insertStageRequest(self, requestDict, pinLifeTime):
        req = "INSERT INTO StageRequests (ReplicaID,RequestID,StageRequestSubmitTime,PinLength) VALUES "
        for requestID, replicaIDs in requestDict.items():
            for replicaID in replicaIDs:
                replicaString = "(%s,'%s',UTC_TIMESTAMP(),%d)," % (replicaID, requestID, pinLifeTime)
                req = "%s %s" % (req, replicaString)
        req = req.rstrip(",")
        res = self._update(req)
        if not res["OK"]:
            gLogger.error(
                "StorageManagementDB.insertStageRequest: Failed to insert to StageRequests table.", res["Message"]
            )
            return res

        for requestID, replicaIDs in requestDict.items():
            for replicaID in replicaIDs:
                # fix, no individual queries
                reqSelect = "SELECT * FROM StageRequests WHERE ReplicaID = %s AND RequestID = '%s';" % (
                    replicaID,
                    requestID,
                )
                resSelect = self._query(reqSelect)
                if not resSelect["OK"]:
                    gLogger.warn(
                        "%s.%s_DB: problem retrieving record: %s. %s"
                        % (self._caller(), "insertStageRequest", reqSelect, resSelect["Message"])
                    )
                else:
                    gLogger.verbose(
                        "%s.%s_DB: inserted StageRequests = %s"
                        % (self._caller(), "insertStageRequest", resSelect["Value"][0])
                    )

        # gLogger.info( "%s_DB: howmany = %s" % ('insertStageRequest',res))

        # gLogger.info( "%s_DB:%s" % ('insertStageRequest',req))
        gLogger.debug(
            "StorageManagementDB.insertStageRequest: Successfully added %s StageRequests with RequestID %s."
            % (res["Value"], requestID)
        )
        return S_OK()

    ####################################################################
    #
    # The state transition of the CacheReplicas from StageSubmitted->Staged
    #

    def setStageComplete(self, replicaIDs):
        # Daniela: FIX wrong PinExpiryTime (84000->86400 seconds = 1 day)

        reqSelect = "SELECT * FROM StageRequests WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
        resSelect = self._query(reqSelect)
        if not resSelect["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "setStageComplete", reqSelect, resSelect["Message"])
            )
            return resSelect

        req = (
            "UPDATE StageRequests SET StageStatus='Staged',StageRequestCompletedTime = UTC_TIMESTAMP(),PinExpiryTime = DATE_ADD(UTC_TIMESTAMP(),INTERVAL ( PinLength / %s ) SECOND) WHERE ReplicaID IN (%s);"
            % (THROTTLING_STEPS, intListToString(replicaIDs))
        )
        res = self._update(req)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.setStageComplete: Failed to set StageRequest completed.", res["Message"])
            return res

        for record in resSelect["Value"]:
            gLogger.verbose("%s.%s_DB: to_update StageRequests =  %s" % (self._caller(), "setStageComplete", record))

        reqSelect1 = "SELECT * FROM StageRequests WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
        resSelect1 = self._query(reqSelect1)
        if not resSelect1["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "setStageComplete", reqSelect1, resSelect1["Message"])
            )

        for record in resSelect1["Value"]:
            gLogger.verbose("%s.%s_DB: updated StageRequests = %s" % (self._caller(), "setStageComplete", record))

        gLogger.debug(
            "StorageManagementDB.setStageComplete: Successfully updated %s StageRequests table with StageStatus=Staged for ReplicaIDs: %s."
            % (res["Value"], replicaIDs)
        )
        return res

    def wakeupOldRequests(self, replicaIDs, retryInterval, connection=False):
        """
        get only StageRequests with StageRequestSubmitTime older than 1 day AND are still not staged
        delete these requests
        reset Replicas with corresponding ReplicaIDs to Status='New'
        """
        try:
            retryInterval = max(retryInterval, 2)
            retryInterval = min(retryInterval, 24)
            retryInterval = int(retryInterval)
        except Exception:
            errorString = "Wrong argument type"
            gLogger.exception(errorString)
            return S_ERROR(errorString)
        if (replicaIDs) > 0:
            req = (
                "SELECT ReplicaID FROM StageRequests WHERE ReplicaID IN (%s) AND StageStatus='StageSubmitted' AND DATE_ADD( StageRequestSubmitTime, INTERVAL %s HOUR ) < UTC_TIMESTAMP();"
                % (intListToString(replicaIDs), retryInterval)
            )
            res = self._query(req)
            if not res["OK"]:
                gLogger.error(
                    "StorageManagementDB.wakeupOldRequests: Failed to select old StageRequests.", res["Message"]
                )
                return res

            old_replicaIDs = [row[0] for row in res["Value"]]

            if len(old_replicaIDs) > 0:
                req = (
                    "UPDATE CacheReplicas SET Status='New',LastUpdate = UTC_TIMESTAMP(), Reason = 'wakeupOldRequests' WHERE ReplicaID in (%s);"
                    % intListToString(old_replicaIDs)
                )
                res = self._update(req, connection)
                if not res["OK"]:
                    gLogger.error(
                        "StorageManagementDB.wakeupOldRequests: Failed to roll CacheReplicas back to Status=New.",
                        res["Message"],
                    )
                    return res

                req = "DELETE FROM StageRequests WHERE ReplicaID in (%s);" % intListToString(old_replicaIDs)
                res = self._update(req, connection)
                if not res["OK"]:
                    gLogger.error("StorageManagementDB.wakeupOldRequests. Problem removing entries from StageRequests.")
                    return res

        return S_OK()

    ####################################################################
    #
    # This code handles the finalization of stage tasks
    #
    # Daniela: useless method
    '''
  def updateStageCompletingTasks(self):
    """ This will select all the Tasks in StageCompleting status and check whether all the associated files are Staged. """
    req = "SELECT TR.TaskID,COUNT(if(R.Status NOT IN ('Staged'),1,NULL)) FROM Tasks AS T, TaskReplicas AS TR, CacheReplicas AS R WHERE T.Status='StageCompleting' AND T.TaskID=TR.TaskID AND TR.ReplicaID=R.ReplicaID GROUP BY TR.TaskID;"
    res = self._query(req)
    if not res['OK']:
      return res
    taskIDs = []
    for taskID,count in res['Value']:
      if int(count) == 0:
        taskIDs.append(taskID)
    if not taskIDs:
      return S_OK(taskIDs)
    req = "UPDATE Tasks SET Status = 'Staged' WHERE TaskID IN (%s);" % intListToString(taskIDs)
    res = self._update(req)
    if not res['OK']:
      return res
    return S_OK(taskIDs)
  '''

    def setTasksDone(self, taskIDs):
        """This will update the status for a list of taskIDs to Done."""
        reqSelect = "SELECT * FROM Tasks WHERE TaskID IN (%s);" % intListToString(taskIDs)
        resSelect = self._query(reqSelect)
        if not resSelect["OK"]:
            gLogger.error(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "setTasksDone", reqSelect, resSelect["Message"])
            )

        req = (
            "UPDATE Tasks SET Status = 'Done', CompleteTime = UTC_TIMESTAMP() WHERE TaskID IN (%s);"
            % intListToString(taskIDs)
        )
        res = self._update(req)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.setTasksDone: Failed to set Tasks status to Done.", res["Message"])
            return res

        for record in resSelect["Value"]:
            gLogger.verbose("%s.%s_DB: to_update Tasks =  %s" % (self._caller(), "setTasksDone", record))
            # fix, no individual queries
        reqSelect1 = "SELECT * FROM Tasks WHERE TaskID IN (%s);" % intListToString(taskIDs)
        resSelect1 = self._query(reqSelect1)
        if not resSelect1["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "setTasksDone", reqSelect1, resSelect1["Message"])
            )
        else:
            for record in resSelect1["Value"]:
                gLogger.verbose("%s.%s_DB: updated Tasks = %s" % (self._caller(), "setTasksDone", record))

        gLogger.debug(
            "StorageManagementDB.setTasksDone: Successfully updated %s Tasks with StageStatus=Done for taskIDs: %s."
            % (res["Value"], taskIDs)
        )
        return res

    def killTasksBySourceTaskID(self, sourceTaskIDs, connection=False):
        """Given SourceTaskIDs (jobs), this will cancel further staging of files for the corresponding tasks.
        The "cancel" is actually removing all stager DB records for these jobs.
        Care must be taken to NOT cancel staging of files that are requested also by other tasks."""
        connection = self.__getConnection(connection)

        # get the TaskIDs
        req = "SELECT TaskID from Tasks WHERE SourceTaskID IN (%s);" % intListToString(sourceTaskIDs)
        res = self._query(req)
        if not res["OK"]:
            gLogger.error(
                "%s.%s_DB: problem retrieving records: %s. %s"
                % (self._caller(), "killTasksBySourceTaskID", req, res["Message"])
            )
        taskIDs = [row[0] for row in res["Value"]]

        # ! Make sure to only cancel file staging for files with no relations with other tasks (jobs) but the killed ones
        if taskIDs:
            req = (
                "SELECT DISTINCT(CR.ReplicaID) FROM TaskReplicas AS TR, CacheReplicas AS CR WHERE TR.TaskID IN (%s) AND CR.Links=1 and TR.ReplicaID=CR.ReplicaID;"
                % intListToString(taskIDs)
            )
            res = self._query(req)
            if not res["OK"]:
                gLogger.error(
                    "%s.%s_DB: problem retrieving records: %s. %s"
                    % (self._caller(), "killTasksBySourceTaskID", req, res["Message"])
                )

            replicaIDs = [row[0] for row in res["Value"]]

            if replicaIDs:
                req = "DELETE FROM StageRequests WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
                res = self._update(req, connection)
                if not res["OK"]:
                    gLogger.error(
                        "%s.%s_DB: problem removing records: %s. %s"
                        % (self._caller(), "killTasksBySourceTaskID", req, res["Message"])
                    )

                req = "DELETE FROM CacheReplicas WHERE ReplicaID in (%s) AND Links=1;" % intListToString(replicaIDs)
                res = self._update(req, connection)
                if not res["OK"]:
                    gLogger.error(
                        "%s.%s_DB: problem removing records: %s. %s"
                        % (self._caller(), "killTasksBySourceTaskID", req, res["Message"])
                    )

            # Finally, remove the Task and TaskReplicas entries.
            res = self.removeTasks(taskIDs, connection)
        return res

    def removeStageRequests(self, replicaIDs, connection=False):
        connection = self.__getConnection(connection)
        req = "DELETE FROM StageRequests WHERE ReplicaID in (%s);" % intListToString(replicaIDs)
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.removeStageRequests. Problem removing entries from StageRequests.")
            return res
        return res

    def removeTasks(self, taskIDs, connection=False):
        """This will delete the entries from the TaskReplicas for the provided taskIDs."""
        connection = self.__getConnection(connection)
        req = "DELETE FROM TaskReplicas WHERE TaskID IN (%s);" % intListToString(taskIDs)
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.removeTasks. Problem removing entries from TaskReplicas.")
            return res
        # gLogger.info( "%s_DB:%s" % ('removeTasks',req))
        reqSelect = "SELECT * FROM Tasks WHERE TaskID IN (%s);" % intListToString(taskIDs)
        resSelect = self._query(reqSelect)
        if not resSelect["OK"]:
            gLogger.error(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "removeTasks", reqSelect, resSelect["Message"])
            )
        else:
            for record in resSelect["Value"]:
                gLogger.verbose("%s.%s_DB: to_delete Tasks =  %s" % (self._caller(), "removeTasks", record))

        req = "DELETE FROM Tasks WHERE TaskID in (%s);" % intListToString(taskIDs)
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.removeTasks. Problem removing entries from Tasks.")
        gLogger.verbose("%s.%s_DB: deleted Tasks" % (self._caller(), "removeTasks"))
        # gLogger.info( "%s_DB:%s" % ('removeTasks',req))
        return res

    def setOldTasksAsFailed(self, daysOld, connection=False):
        """
        Set Tasks older than "daysOld" number of days to Failed
        These tasks have already been retried every day for staging
        """
        req = "UPDATE Tasks SET Status='Failed' WHERE DATE_ADD(SubmitTime, INTERVAL %s DAY ) < UTC_TIMESTAMP();" % (
            daysOld
        )
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.setOldTasksAsFailed. Problem setting old Tasks to Failed.")
            return res
        return res

    def getCacheReplicasSummary(self, connection=False):
        """
        Reports breakdown of file number/size in different staging states across storage elements
        """
        connection = self.__getConnection(connection)
        req = "SELECT DISTINCT(Status),SE,COUNT(*),sum(size)/(1024*1024*1024) FROM CacheReplicas GROUP BY Status,SE;"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error("StorageManagementDB.getCacheReplicasSummary failed.")
            return res

        resSummary = {}
        i = 1
        for status, se, numFiles, sumFiles in res["Value"]:
            resSummary[i] = {"Status": status, "SE": se, "NumFiles": long(numFiles), "SumFiles": float(sumFiles)}
            i += 1
        return S_OK(resSummary)

    def removeUnlinkedReplicas(self, connection=False):
        """This will remove Replicas from the CacheReplicas that are not associated to any Task.
        If the Replica has been Staged,
        wait until StageRequest.PinExpiryTime and remove the StageRequest and CacheReplicas entries
        """
        connection = self.__getConnection(connection)
        # First, check if there is a StageRequest and PinExpiryTime has arrived
        req = "select SR.ReplicaID from CacheReplicas CR,StageRequests SR WHERE CR.Links = 0 and CR.ReplicaID=SR.ReplicaID group by SR.ReplicaID HAVING max(SR.PinExpiryTime) < UTC_TIMESTAMP();"
        # req = "SELECT ReplicaID from CacheReplicas WHERE Links = 0;"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error(
                "StorageManagementDB.removeUnlinkedReplicas. Problem selecting entries from CacheReplicas where Links = 0."
            )
            return res

        replicaIDs = [row[0] for row in res["Value"]]
        # Look for Failed CacheReplicas which are not associated to any Task. These have no PinExpiryTime in StageRequests
        # as they were not staged successfully (for various reasons), even though
        # a staging request had been submitted
        req = "SELECT ReplicaID FROM CacheReplicas WHERE Links = 0 AND Status = 'Failed';"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error(
                "StorageManagementDB.removeUnlinkedReplicas. Problem selecting entries from CacheReplicas where Links = 0 AND Status=Failed."
            )
        else:
            replicaIDs.extend([row[0] for row in res["Value"]])

        if replicaIDs:
            # Removed the entries from the StageRequests table that are expired
            reqSelect = "SELECT * FROM StageRequests WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
            resSelect = self._query(reqSelect)
            if not resSelect["OK"]:
                gLogger.warn(
                    "%s.%s_DB: problem retrieving record: %s. %s"
                    % (self._caller(), "removeUnlinkedReplicas", reqSelect, resSelect["Message"])
                )
            else:
                for record in resSelect["Value"]:
                    gLogger.verbose(
                        "%s.%s_DB: to_delete StageRequests = %s" % (self._caller(), "removeUnlinkedReplicas", record)
                    )

            req = "DELETE FROM StageRequests WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
            res = self._update(req, connection)
            if not res["OK"]:
                gLogger.error("StorageManagementDB.removeUnlinkedReplicas. Problem deleting from StageRequests.")
                return res
            gLogger.verbose("%s.%s_DB: deleted StageRequests" % (self._caller(), "removeUnlinkedReplicas"))

            gLogger.debug(
                "StorageManagementDB.removeUnlinkedReplicas: Successfully removed %s StageRequests entries for ReplicaIDs: %s."
                % (res["Value"], replicaIDs)
            )

        # Second look for CacheReplicas for which there is no entry in StageRequests
        req = "SELECT ReplicaID FROM CacheReplicas WHERE Links = 0 AND ReplicaID NOT IN ( SELECT DISTINCT( ReplicaID ) FROM StageRequests )"
        res = self._query(req, connection)
        if not res["OK"]:
            gLogger.error(
                "StorageManagementDB.removeUnlinkedReplicas. Problem selecting entries from CacheReplicas where Links = 0."
            )
        else:
            replicaIDs.extend([row[0] for row in res["Value"]])

        if not replicaIDs:
            return S_OK()

        # Now delete all CacheReplicas
        reqSelect = "SELECT * FROM CacheReplicas WHERE ReplicaID IN (%s);" % intListToString(replicaIDs)
        resSelect = self._query(reqSelect)
        if not resSelect["OK"]:
            gLogger.warn(
                "%s.%s_DB: problem retrieving record: %s. %s"
                % (self._caller(), "removeUnlinkedReplicas", reqSelect, resSelect["Message"])
            )
        else:
            for record in resSelect["Value"]:
                gLogger.verbose(
                    "%s.%s_DB: to_delete CacheReplicas =  %s" % (self._caller(), "removeUnlinkedReplicas", record)
                )

        req = "DELETE FROM CacheReplicas WHERE ReplicaID IN (%s) AND Links= 0;" % intListToString(replicaIDs)
        res = self._update(req, connection)
        if res["OK"]:
            gLogger.verbose("%s.%s_DB: deleted CacheReplicas" % (self._caller(), "removeUnlinkedReplicas"))
            gLogger.debug(
                "StorageManagementDB.removeUnlinkedReplicas: Successfully removed %s CacheReplicas entries for ReplicaIDs: %s."
                % (res["Value"], replicaIDs)
            )
        else:
            gLogger.error("StorageManagementDB.removeUnlinkedReplicas. Problem removing entries from CacheReplicas.")
        return res



""" Test the ProxyPlugin class"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import unittest
import mock
from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin import ProxyPlugin

__RCSID__ = "$Id $"


def mock_getProxyInfo():
    return S_OK(
        {
            "VOMS": ["/lhcb/Role=user"],
            "chain": "[/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=chaen/CN=123456/CN=Christophe Haen]",
            "group": "lhcb_user",
            "groupProperties": ["NormalUser", "SuperProperty"],
            "hasVOMS": True,
            "identity": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=chaen/CN=123456/CN=Christophe Haen",
            "isLimitedProxy": False,
            "isProxy": True,
            "issuer": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=chaen/CN=123456/CN=Christophe Haen/CN=proxy",
            "path": "/tmp/x509up_u12345",
            "secondsLeft": 86026,
            "subject": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=chaen/CN=123456/CN=Christophe Haen/CN=proxy/CN=proxy",  # noqa # pylint: disable=line-too-long
            "username": "chaen",
            "validDN": True,
            "validGroup": True,
        }
    )


def mock_getNoProxyInfo():
    return S_ERROR("No proxy")


class TestProxyPlugin(unittest.TestCase):
    """Test the FilenamePlugin class"""

    @mock.patch("DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin.getProxyInfo", side_effect=mock_getProxyInfo)
    def test_01_username(self, _mockProxyInfo):
        """Testing username attribute"""

        # the username is chaen
        self.assertTrue(not ProxyPlugin("username.in(toto)").eval())
        self.assertTrue(ProxyPlugin("username.not_in(toto)").eval())
        self.assertTrue(ProxyPlugin("username.in(toto, chaen)").eval())
        self.assertTrue(not ProxyPlugin("username.not_in(toto, chaen)").eval())

        # Testing some more formating with spaces and quotes

        self.assertTrue(not ProxyPlugin("username.in( toto )").eval())
        self.assertTrue(ProxyPlugin('username.not_in("toto")').eval())
        self.assertTrue(ProxyPlugin('username.in(toto,"chaen")').eval())
        self.assertTrue(not ProxyPlugin("username.not_in('toto' ,chaen)").eval())

    @mock.patch("DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin.getProxyInfo", side_effect=mock_getProxyInfo)
    def test_02_group(self, _mockProxyInfo):
        """Testing group attribute"""

        # the group is lhcb_user
        self.assertTrue(not ProxyPlugin("group.in(toto)").eval())
        self.assertTrue(ProxyPlugin("group.not_in(toto)").eval())
        self.assertTrue(ProxyPlugin("group.in(toto, lhcb_user)").eval())
        self.assertTrue(not ProxyPlugin("group.not_in(toto, lhcb_user)").eval())

        # Testing some more formating with spaces and quotes

        self.assertTrue(not ProxyPlugin("group.in( toto )").eval())
        self.assertTrue(ProxyPlugin('group.not_in("toto")').eval())
        self.assertTrue(ProxyPlugin('group.in(toto,"lhcb_user")').eval())
        self.assertTrue(not ProxyPlugin("group.not_in('toto' ,lhcb_user)").eval())

    @mock.patch("DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin.getProxyInfo", side_effect=mock_getProxyInfo)
    def test_03_property(self, _mockProxyInfo):
        """Testing property attribute"""

        # the properties are 'NormalUser', 'SuperProperty'
        self.assertTrue(not ProxyPlugin("property.has(toto)").eval())
        self.assertTrue(ProxyPlugin("property.has_not(toto)").eval())
        self.assertTrue(ProxyPlugin("property.has(SuperProperty)").eval())
        self.assertTrue(not ProxyPlugin("property.has_not(SuperProperty)").eval())

    @mock.patch("DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin.getProxyInfo", side_effect=mock_getProxyInfo)
    def test_04_voms(self, _mockProxyInfo):
        """Testing voms attribute"""

        # the voms role is /lhcb/Role=user
        self.assertTrue(not ProxyPlugin("voms.has(toto)").eval())
        self.assertTrue(ProxyPlugin("voms.has_not(toto)").eval())
        self.assertTrue(ProxyPlugin("voms.has(/lhcb/Role->user)").eval())
        self.assertTrue(not ProxyPlugin("voms.has_not(/lhcb/Role->user)").eval())

    @mock.patch("DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin.getProxyInfo", side_effect=mock_getProxyInfo)
    def test_05_errors(self, _mockProxyInfo):
        """Testing errors handling"""

        # Non existing attribute
        with self.assertRaises(RuntimeError):
            ProxyPlugin("boom.boom(something)").eval()

        # Non matching predicate with the attribute
        with self.assertRaises(RuntimeError):
            ProxyPlugin("username.boom(something)").eval()

        with self.assertRaises(RuntimeError):
            ProxyPlugin("group.boom(something)").eval()

        with self.assertRaises(RuntimeError):
            ProxyPlugin("property.boom(something)").eval()

        with self.assertRaises(RuntimeError):
            ProxyPlugin("voms.boom(something)").eval()

        # impossible parsing
        with self.assertRaises(AttributeError):
            ProxyPlugin("cannotbeparsed").eval()

    @mock.patch("DIRAC.Resources.Catalog.ConditionPlugins.ProxyPlugin.getProxyInfo", side_effect=mock_getNoProxyInfo)
    def test_05_withoutProxy(self, _mockProxyInfo):
        """Testing without proxy, everything should be False"""

        # A priori, not both of them can give the same result
        # but since there is no proxy, it should !
        self.assertFalse(ProxyPlugin("username.in(toto)").eval())
        self.assertFalse(ProxyPlugin("username.not_in(toto)").eval())


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestProxyPlugin)

    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
########################################################################
# File :    dirac-dms-add-file
# Author :  Stuart Paterson
########################################################################
"""
Upload a file to the grid storage and register it in the File Catalog

Usage:
  dirac-dms-add-file [options] ... LFN Path SE [GUID]

Arguments:
  LFN:      Logical File Name
  Path:     Local path to the file
  SE:       DIRAC Storage Element
  GUID:     GUID to use in the registration (optional)

**OR**

Usage:
  dirac-dms-add-file [options] ... LocalFile ...

Arguments:
  LocalFile: Path to local file containing all the above, i.e.::

  lfn1 localfile1 SE [GUID1]
  lfn2 localfile2 SE [GUID2]

Example:
  $ dirac-dms-add-file LFN:/formation/user/v/vhamar/Example.txt Example.txt DIRAC-USER
  {'Failed': {},
   'Successful': {'/formationes/user/v/vhamar/Example.txt': {'put': 0.70791220664978027,
                                                             'register': 0.61061787605285645}}}
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
from DIRAC import S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

overwrite = False


def setOverwrite(arg):
    global overwrite
    overwrite = True
    return S_OK()


def getDict(item_list):
    """
    From the input list, populate the dictionary
    """
    lfn_dict = {}
    lfn_dict["lfn"] = item_list[0].replace("LFN:", "").replace("lfn:", "")
    lfn_dict["localfile"] = item_list[1]
    lfn_dict["SE"] = item_list[2]
    guid = None
    if len(item_list) > 3:
        guid = item_list[3]
    lfn_dict["guid"] = guid
    return lfn_dict


@Script()
def main():
    global overwrite
    Script.registerSwitch("f", "force", "Force overwrite of existing file", setOverwrite)
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if len(args) < 1 or len(args) > 4:
        Script.showHelp(exitCode=1)

    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from DIRAC import gLogger
    import DIRAC

    exitCode = 0

    lfns = []
    if len(args) == 1:
        inputFileName = args[0]
        if os.path.exists(inputFileName):
            inputFile = open(inputFileName, "r")
            for line in inputFile:
                line = line.rstrip()
                items = line.split()
                items[0] = items[0].replace("LFN:", "").replace("lfn:", "")
                lfns.append(getDict(items))
            inputFile.close()
        else:
            gLogger.error("Error: LFN list '%s' missing." % inputFileName)
            exitCode = 4
    else:
        lfns.append(getDict(args))

    dm = DataManager()
    for lfn in lfns:
        if not os.path.exists(lfn["localfile"]):
            gLogger.error("File %s must exist locally" % lfn["localfile"])
            exitCode = 1
            continue
        if not os.path.isfile(lfn["localfile"]):
            gLogger.error("%s is not a file" % lfn["localfile"])
            exitCode = 2
            continue

        gLogger.notice("\nUploading %s" % lfn["lfn"])
        res = dm.putAndRegister(lfn["lfn"], lfn["localfile"], lfn["SE"], lfn["guid"], overwrite=overwrite)
        if not res["OK"]:
            exitCode = 3
            gLogger.error("Error: failed to upload %s to %s: %s" % (lfn["lfn"], lfn["SE"], res))
            continue
        else:
            gLogger.notice("Successfully uploaded file to %s" % lfn["SE"])

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :   dirac-configuration-cli
# Author : Adria Casajus
########################################################################
"""
Command line interface to DIRAC Configuration Server
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.ConfigurationSystem.Client.CSCLI import CSCLI


@Script()
def main():
    Script.localCfg.addDefaultEntry("LogLevel", "fatal")
    Script.parseCommandLine()

    CSCLI().start()


if __name__ == "__main__":
    main()



"""
    Extremely simple utility class to send mails
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import socket

from smtplib import SMTP, SMTP_SSL
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from getpass import getuser

from DIRAC import gLogger, S_OK, S_ERROR

__RCSID__ = "$Id$"


class Mail(object):
    def __init__(self):
        self._subject = ""
        self._message = ""
        self._mailAddress = ""
        self._html = False
        self._fromAddress = getuser() + "@" + socket.getfqdn()
        self._attachments = []
        self.esmtp_features = {}
        self._smtpPtcl = None
        self._smtpHost = None
        self._smtpPort = None
        self._smtpLogin = None
        self._smtpPasswd = None

    def _create(self, addresses):
        """create a mail object

        :param list addresses: addresses

        :return: S_OK(object)/S_ERROR() -- contain MIMEMultipart object
        """
        if not isinstance(addresses, list):
            addresses = [addresses]

        if not self._mailAddress:
            gLogger.warn("No mail address was provided. Mail not sent.")
            return S_ERROR("No mail address was provided. Mail not sent.")

        if not self._message:
            gLogger.warn("Message body is empty")
            if not self._subject:
                gLogger.warn("Subject and body empty. Mail not sent")
                return S_ERROR("Subject and body empty. Mail not sent")

        if self._html:
            mail = MIMEText(self._message, "html")
        else:
            mail = MIMEText(self._message, "plain")

        msg = MIMEMultipart()
        msg.attach(mail)

        msg["Subject"] = self._subject
        msg["From"] = self._fromAddress
        msg["To"] = ", ".join(addresses)

        for attachment in self._attachments:
            try:
                with open(attachment, "rb") as fil:
                    part = MIMEApplication(fil.read(), Name=os.path.basename(attachment))

                    part["Content-Disposition"] = 'attachment; filename="%s"' % os.path.basename(attachment)
                    msg.attach(part)
            except IOError as e:
                gLogger.exception("Could not attach %s" % attachment, lException=e)

        return S_OK(msg)

    def _send(self, msg=None):
        """send a single email message. If msg is in input, it is expected to be of email type, otherwise it will create it.

        :param object msg: MIMEMultipart object

        :return: S_OK()/S_ERROR()
        """

        if msg is None:
            addresses = self._mailAddress
            if isinstance(self._mailAddress, six.string_types):
                addresses = self._mailAddress.split(", ")

            result = self._create(addresses)
            if not result["OK"]:
                return result
            msg = result["Value"]

        if self._smtpPtcl == "SSL":
            smtp = SMTP_SSL()
        else:
            smtp = SMTP()
        smtp.set_debuglevel(0)
        try:
            connParams = {}
            if self._smtpHost:
                connParams["host"] = self._smtpHost
            if self._smtpPort:
                connParams["port"] = int(self._smtpPort)
            smtp.connect(**connParams)
            smtp.ehlo_or_helo_if_needed()
            if self._smtpPtcl == "TLS":
                smtp.starttls()
            if self._smtpLogin and self._smtpPasswd:
                smtp.login(self._smtpLogin, self._smtpPasswd)
            smtp.ehlo_or_helo_if_needed()
            smtp.sendmail(self._fromAddress, addresses, msg.as_string())
        except Exception as x:
            return S_ERROR("Sending mail failed %s" % str(x))

        smtp.quit()
        return S_OK("The mail was successfully sent")

    def __eq__(self, other):
        """Comparing an email object to another"""
        if isinstance(other, Mail):
            if self.__dict__ == other.__dict__:
                return True

        return False

    def __hash__(self):
        """Comparing for sets"""
        return hash(self._subject + self._message + self._fromAddress + self._mailAddress)



"""
ComponentMonitoring type used to monitor DIRAC components.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.MonitoringSystem.Client.Types.BaseType import BaseType

__RCSID__ = "$Id$"


class ComponentMonitoring(BaseType):
    """
    .. class:: ComponentMonitoring
    """

    def __init__(self):
        """c'tor

        :param self: self reference
        """

        super(ComponentMonitoring, self).__init__()

        self.keyFields = ["host", "component", "pid", "status", "componentType", "componentLocation"]

        self.monitoringFields = [
            "runningTime",
            "memoryUsage",
            "threads",
            "cpuPercentage",
            "Connections",
            "PendingQueries",
            "ActiveQueries",
            "RunningThreads",
            "MaxFD",
            "ServiceResponseTime",
            "cycleDuration",
            "cycles",
        ]

        self.addMapping(
            {
                "host": {"type": "keyword"},
                "component": {"type": "keyword"},
                "status": {"type": "keyword"},
                "componentType": {"type": "keyword"},
                "componentLocation": {"type": "keyword"},
            }
        )

        self.dataToKeep = 86400 * 30  # we need to define...

        self.period = "month"
        self.checkType()



"""
  This module manages is a client for the TS and manages the transformations associated to the productions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

import json

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient


class ProdTransManager(object):
    def __init__(self):
        self.transClient = TransformationClient()
        self.prodClient = ProductionClient()

    def deleteTransformations(self, transIDs):
        """Delete given transformations from the TS

        :param list transIDs: a list of Transformation IDs
        """
        gLogger.notice("Deleting transformations %s from the TS" % transIDs)

        for transID in transIDs:
            res = self.transClient.deleteTransformation(transID)
            if not res["OK"]:
                return res

        return S_OK()

    def deleteProductionTransformations(self, prodID):
        """Delete the production transformations from the TS

        :param int prodID: the ProductionID
        """
        res = self.prodClient.getProductionTransformations(prodID)
        if res["OK"]:
            transList = res["Value"]

        gLogger.notice("Deleting production transformations %s from the TS" % transList)

        for trans in transList:
            transID = trans["TransformationID"]
            res = self.transClient.deleteTransformation(transID)
            if not res["OK"]:
                gLogger.error(res["Message"])

        return S_OK()

    def addTransformationStep(self, stepID, prodID):
        """Add the transformation step to the TS

        :param int stepID: the production step ID
        :param int prodID: the production ID
        :return:
        """

        res = self.prodClient.getProductionStep(stepID)
        if not res["OK"]:
            return res
        prodStep = res["Value"]

        gLogger.notice("Add step %s to production %s" % (prodStep[0], prodID))

        stepDesc = prodStep[2]
        stepLongDesc = prodStep[3]
        stepBody = prodStep[4]
        stepType = prodStep[5]
        stepPlugin = prodStep[6]
        stepAgentType = prodStep[7]
        stepGroupsize = prodStep[8]
        stepInputquery = json.loads(prodStep[9])
        stepOutputquery = json.loads(prodStep[10])

        stepName = "%08d" % prodID + "_" + prodStep[1]

        res = self.transClient.addTransformation(
            stepName,
            stepDesc,
            stepLongDesc,
            stepType,
            stepPlugin,
            stepAgentType,
            "",
            groupSize=stepGroupsize,
            body=stepBody,
            inputMetaQuery=stepInputquery,
            outputMetaQuery=stepOutputquery,
        )

        if not res["OK"]:
            return S_ERROR(res["Message"])

        return S_OK(res["Value"])

    def executeActionOnTransformations(self, prodID, action):
        """Wrapper to start/stop/clean the transformations of a production

        :param int prodID: the production ID
        :param str action: it can be start/stop/clean
        """

        # Check if there is any action to do
        if not action:
            return S_OK()

        # Get the transformations of the production
        res = self.prodClient.getProductionTransformations(prodID)
        if not res["OK"]:
            return res

        transList = res["Value"]
        method = getattr(self.transClient, action)
        gLogger.notice("Executing action %s to %s" % (action, transList))

        # Execute the action on each transformation
        for trans in transList:
            transID = trans["TransformationID"]
            res = method(transID)
            if not res["OK"]:
                return res

        return S_OK()



#!/usr/bin/env python

# spawns 4 processes (in theory on 4 processors...)

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from multiprocessing import Process, Queue, current_process


def do_sum(q, l):
    q.put(sum(l))
    proc_name = current_process().name
    print(proc_name)  # pylint: disable=print-statement


def main():
    my_list = list(range(20000000))

    q = Queue()

    p1 = Process(target=do_sum, args=(q, my_list[:5000000]))
    p2 = Process(target=do_sum, args=(q, my_list[5000000:10000000]))
    p3 = Process(target=do_sum, args=(q, my_list[10000000:15000000]))
    p4 = Process(target=do_sum, args=(q, my_list[15000000:]))
    p1.start()
    p2.start()
    p3.start()
    p4.start()

    r1 = q.get()
    r2 = q.get()
    r3 = q.get()
    r4 = q.get()
    q.close()
    q.join_thread()

    p1.join()
    p2.join()
    p3.join()
    p4.join()

    print(r1 + r2 + r3 + r4)  # pylint: disable=print-statement


if __name__ == "__main__":
    main()



""" unit test for Watchdog.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
import os
from mock import MagicMock

# sut
from DIRAC.WorkloadManagementSystem.JobWrapper.Watchdog import Watchdog

mock_exeThread = MagicMock()
mock_spObject = MagicMock()


def test_calibrate():
    pid = os.getpid()
    wd = Watchdog(pid, mock_exeThread, mock_spObject, 5000)
    res = wd.calibrate()
    assert res["OK"] is True


def test__performChecks():
    pid = os.getpid()
    wd = Watchdog(pid, mock_exeThread, mock_spObject, 5000)

    res = wd.calibrate()
    assert res["OK"] is True
    res = wd._performChecks()
    assert res["OK"] is True


def test__performChecksFull():
    pid = os.getpid()
    wd = Watchdog(pid, mock_exeThread, mock_spObject, 5000)
    wd.testCPULimit = 1
    wd.testMemoryLimit = 1

    res = wd.calibrate()
    assert res["OK"] is True
    res = wd._performChecks()
    assert res["OK"] is True



########################################################################
# File :   CLI.py
# Author : Andrei Tsaregorodtsev
########################################################################

""" CLI is the base class for all the DIRAC consoles ( CLIs ). It contains
    several utilities and signal handlers of general purpose.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
import cmd
import sys
import os
import signal

from DIRAC import gLogger

gColors = {"red": 1, "green": 2, "yellow": 3, "blue": 4}


def colorEnabled():
    return os.environ.get("TERM") in ("xterm", "xterm-color")


def colorize(text, color):
    """Return colorized text"""
    if not colorEnabled():
        return text

    startCode = "\033[;3"
    endCode = "\033[0m"
    if isinstance(color, six.integer_types):
        return "%s%sm%s%s" % (startCode, color, text, endCode)
    try:
        return "%s%sm%s%s" % (startCode, gColors[color], text, endCode)
    except Exception:
        return text


class CLI(cmd.Cmd):
    def __init__(self):

        cmd.Cmd.__init__(self)
        self.indentSpace = 20
        self._initSignals()

    def _handleSignal(self, sig, frame):

        print("\nReceived signal", sig, ", exiting ...")
        self.do_quit(self)

    def _initSignals(self):
        """
        Registers signal handlers
        """
        for sigNum in (signal.SIGINT, signal.SIGQUIT, signal.SIGKILL, signal.SIGTERM):
            try:
                signal.signal(sigNum, self._handleSignal)
            except Exception:
                pass

    def _errMsg(self, errMsg):
        """
        Print out a colorized error log message

        :param str errMsg: error message string
        :return: nothing
        """
        gLogger.error("%s %s" % (colorize("[ERROR]", "red"), errMsg))

    def emptyline(self):
        pass

    def do_exit(self, args):
        """Exit the shell.

        usage: exit
        """
        self.do_quit(self)

    def do_quit(self, args):
        """Exit the shell.

        usage: quit
        """
        gLogger.notice("")
        sys.exit(0)

    def do_EOF(self, args):
        """Handler for EOF ( Ctrl D ) signal - perform quit"""
        self.do_quit(args)

    def do_execfile(self, args):
        """Execute a series of CLI commands from a given file

        usage:

          execfile <filename>
        """

        argss = args.split()
        fname = argss[0]
        if not os.path.exists(fname):
            print("Error: File not found %s" % fname)
            return
        with open(fname, "r") as input_cmd:
            contents = input_cmd.readlines()
        for line in contents:
            try:
                gLogger.notice("\n--> Executing %s\n" % line)
                self.onecmd(line)
            except Exception as error:
                self._errMsg(str(error))
                break
        return

    def printPair(self, key, value, separator=":"):
        valueList = value.split("\n")
        print("%s%s%s %s" % (key, " " * (self.indentSpace - len(key)), separator, valueList[0].strip()))
        for valueLine in valueList[1:-1]:
            print("%s  %s" % (" " * self.indentSpace, valueLine.strip()))

    def do_help(self, args):
        """
        Shows help information
            Usage: help <command>
            If no command is specified all commands are shown
        """
        if len(args) == 0:
            print("\nAvailable commands:\n")
            attrList = sorted(dir(self))
            for attribute in attrList:
                if attribute.startswith("do_"):
                    self.printPair(attribute[3:], getattr(self, attribute).__doc__[1:])
                    print("")
        else:
            command = args.split()[0].strip()
            try:
                obj = getattr(self, "do_%s" % command)
            except Exception:
                print("There's no such %s command" % command)
                return
            self.printPair(command, obj.__doc__[1:])



"""
   CloudEndpoint is a base class for the clients used to connect to different
   cloud providers
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

__RCSID__ = "$Id$"

import os
import ssl
import time

from libcloud import security
from libcloud.compute.types import Provider
from libcloud.compute.providers import get_driver
from libcloud.common.exceptions import BaseHTTPError

# DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import makeGuid

from DIRAC.Resources.Cloud.Endpoint import Endpoint
from DIRAC.Resources.Cloud.Utilities import STATE_MAP


class CloudEndpoint(Endpoint):
    """CloudEndpoint base class"""

    def __init__(self, parameters=None):
        super(CloudEndpoint, self).__init__(parameters=parameters)
        # logger
        self.log = gLogger.getSubLogger("CloudEndpoint")
        self.valid = False
        result = self.initialize()
        if result["OK"]:
            self.log.debug("CloudEndpoint created and validated")
            self.valid = True

    def initialize(self):

        # Relax security
        security.SSL_VERSION = ssl.PROTOCOL_SSLv23
        security.VERIFY_SSL_CERT = False

        # Variables needed to contact the service
        connDict = {}
        for var in [
            "ex_domain_name",
            "ex_force_auth_url",
            "ex_force_service_region",
            "ex_force_auth_version",
            "ex_tenant_name",
            "ex_keyname",
            "ex_voms_proxy",
        ]:
            if var in self.parameters:
                connDict[var] = self.parameters[var]

        username = self.parameters.get("User")
        password = self.parameters.get("Password")

        for key in connDict:
            self.log.info("%s: %s" % (key, connDict[key]))

        # get cloud driver
        providerName = self.parameters.get("Provider", "OPENSTACK").upper()
        providerCode = getattr(Provider, providerName)
        self.driverClass = get_driver(providerCode)

        self.__driver = self.driverClass(username, password, **connDict)

        return self.__checkConnection()

    def __checkConnection(self):
        """
        Checks connection status by trying to list the images.

        :return: S_OK | S_ERROR
        """
        try:
            _result = self.__driver.list_images()
        except Exception as errmsg:
            return S_ERROR(errmsg)

        return S_OK()

    def __getImageByName(self, imageName):
        """
        Given the imageName, returns the current image object from the server.

        :Parameters:
          **imageName** - `string`
            imageName as stored on the OpenStack image repository ( glance )

        :return: S_OK( image ) | S_ERROR
        """
        try:
            images = self.__driver.list_images()
        except Exception as errmsg:
            return S_ERROR(errmsg)

        image = None
        for im in images:
            if im.name == imageName:
                image = im
                break

        if image is None:
            return S_ERROR("Image %s not found" % imageName)

        return S_OK(image)

    def __getFlavorByName(self, flavorName):
        """
        Given the flavorName, returns the current flavor object from the server.

        :Parameters:
          **flavorName** - `string`
            flavorName as stored on the OpenStack service

        :return: S_OK( flavor ) | S_ERROR
        """
        try:
            flavors = self.__driver.list_sizes()
        except Exception as errmsg:
            return S_ERROR(errmsg)

        flavor = None
        for fl in flavors:
            if fl.name == flavorName:
                flavor = fl

        if flavor is None:
            return S_ERROR("Flavor %s not found" % flavorName)

        return S_OK(flavor)

    def __getSecurityGroups(self, securityGroupNames=None):
        """
        Given the securityGroupName, returns the current security group object from the server.

        :Parameters:
          **securityGroupName** - `string`
            securityGroupName as stored on the OpenStack service

        :return: S_OK( securityGroup ) | S_ERROR
        """

        if not securityGroupNames:
            securityGroupNames = []
        elif not isinstance(securityGroupNames, list):
            securityGroupNames = [securityGroupNames]

        if "default" not in securityGroupNames:
            securityGroupNames.append("default")

        try:
            secGroups = self.__driver.ex_list_security_groups()
        except Exception as errmsg:
            return S_ERROR(errmsg)

        return S_OK([secGroup for secGroup in secGroups if secGroup.name in securityGroupNames])

    def createInstances(self, vmsToSubmit):
        outputDict = {}

        for nvm in range(vmsToSubmit):
            instanceID = makeGuid()[:8]
            createPublicIP = "ipPool" in self.parameters
            result = self.createInstance(instanceID, createPublicIP)
            if result["OK"]:
                node, publicIP = result["Value"]
                self.log.debug("Created VM instance %s/%s with publicIP %s" % (node.id, instanceID, publicIP))
                nodeDict = {}
                nodeDict["PublicIP"] = publicIP
                nodeDict["InstanceID"] = instanceID
                nodeDict["NumberOfProcessors"] = self.flavor.vcpus
                nodeDict["RAM"] = self.flavor.ram
                nodeDict["DiskSize"] = self.flavor.disk
                nodeDict["Price"] = self.flavor.price
                outputDict[node.id] = nodeDict
            else:
                break

        if not outputDict:
            # Submission failed
            return result

        return S_OK(outputDict)

    def createInstance(self, instanceID="", createPublicIP=True):
        """
        This creates a VM instance for the given boot image
        and creates a context script, taken the given parameters.
        Successful creation returns instance VM

        Boots a new node on the OpenStack server defined by self.endpointConfig. The
        'personality' of the node is done by self.imageConfig. Both variables are
        defined on initialization phase.

        The node name has the following format:
        <bootImageName><contextMethod><time>

        It boots the node. If IPpool is defined on the imageConfiguration, a floating
        IP is created and assigned to the node.

        :return: S_OK( ( nodeID, publicIP ) ) | S_ERROR
        """

        if not instanceID:
            instanceID = makeGuid()[:8]

        self.parameters["VMUUID"] = instanceID
        self.parameters["VMType"] = self.parameters.get("CEType", "OpenStack")

        createNodeDict = {}

        # Get the image object
        if "ImageID" in self.parameters:
            try:
                image = self.__driver.get_image(self.parameters["ImageID"])
            except BaseHTTPError as err:
                if err.code == 404:
                    # Image not found
                    return S_ERROR("Image with ID %s not found" % self.parameters["ImageID"])
                return S_ERROR("Failed to get image for ID %s (%s)" % (self.parameters["ImageID"], str(err)))
        elif "ImageName" in self.parameters:
            result = self.__getImageByName(self.parameters["ImageName"])
            if not result["OK"]:
                return result
            image = result["Value"]
        else:
            return S_ERROR("No image specified")
        createNodeDict["image"] = image

        # Get the flavor object
        if "FlavorID" in self.parameters and "FlavorName" not in self.parameters:
            result = self.__getFlavorByName(self.parameters["FlavorName"])
            if not result["OK"]:
                return result
            flavor = result["Value"]
        elif "FlavorID" in self.parameters:
            flavor = self.__driver.ex_get_size(self.parameters["FlavorID"])
        else:
            return S_ERROR("No flavor specified")
        self.flavor = flavor

        createNodeDict["size"] = flavor

        # Get security groups
        # if 'ex_security_groups' in self.parameters:
        #  result = self.__getSecurityGroups( self.parameters['ex_security_groups'] )
        #  if not result[ 'OK' ]:
        #    self.log.error( result[ 'Message' ] )
        #    return result
        #  self.parameters['ex_security_groups'] = result[ 'Value' ]

        result = self._createUserDataScript()
        if not result["OK"]:
            return result

        createNodeDict["ex_userdata"] = result["Value"]

        # Optional node contextualization parameters
        for param in ["ex_metadata", "ex_pubkey_path", "ex_keyname", "ex_config_drive"]:
            if param in self.parameters:
                createNodeDict[param] = self.parameters[param]

        createNodeDict["name"] = "DIRAC_%s" % instanceID

        # createNodeDict['ex_config_drive'] = True

        self.log.verbose("Creating node:")
        for key, value in createNodeDict.items():
            self.log.verbose("%s: %s" % (key, value))

        if "networks" in self.parameters:
            result = self.getVMNetwork()
            if not result["OK"]:
                return result
            createNodeDict["networks"] = result["Value"]
        if "keyname" in self.parameters:
            createNodeDict["ex_keyname"] = self.parameters["keyname"]

        if "availability_zone" in self.parameters:
            createNodeDict["ex_availability_zone"] = self.parameters["availability_zone"]

        # Create the VM instance now
        try:
            vmNode = self.__driver.create_node(**createNodeDict)

        except Exception as errmsg:
            self.log.error("Exception in driver.create_node", errmsg)
            return S_ERROR(errmsg)

        publicIP = None
        if createPublicIP:

            # Wait until the node is running, otherwise getting public IP fails
            try:
                self.__driver.wait_until_running([vmNode], timeout=600)
                result = self.assignFloatingIP(vmNode)
                if result["OK"]:
                    publicIP = result["Value"]
                else:
                    vmNode.destroy()
                    return result
            except Exception as exc:
                self.log.debug("Failed to wait node running %s" % str(exc))
                vmNode.destroy()
                return S_ERROR("Failed to wait until the node is Running")

        return S_OK((vmNode, publicIP))

    def getVMNodes(self):
        """Get all the nodes on the endpoint

        :return: S_OK(list of Node) / S_ERROR
        """

        try:
            nodes = self.__driver.list_nodes()
        except Exception as errmsg:
            return S_ERROR(errmsg)

        return S_OK(nodes)

    def getVMNode(self, nodeID):
        """
        Given a Node ID, returns all its configuration details on a
        libcloud.compute.base.Node object.

        :Parameters:
          **nodeID** - `string`
            openstack node id ( not uuid ! )

        :return: S_OK( Node ) | S_ERROR
        """

        try:
            node = self.__driver.ex_get_node_details(nodeID)
        except Exception as errmsg:
            # Let's if the node is in the list of available nodes
            result = self.getVMNodes()
            if not result["OK"]:
                return S_ERROR("Failed to get nodes")
            nodeList = result["Value"]
            for nd in nodeList:
                if nd.id == nodeID:
                    # Let's try again
                    try:
                        node = self.__driver.ex_get_node_details(nodeID)
                        break
                    except Exception as exc:
                        return S_ERROR("Failed to get node details %s" % str(exc))
            node = None

        return S_OK(node)

    def getVMStatus(self, nodeID):
        """
        Get the status for a given node ID. libcloud translates the status into a digit
        from 0 to 4 using a many-to-one relation ( ACTIVE and RUNNING -> 0 ), which
        means we cannot undo that translation. It uses an intermediate states mapping
        dictionary, SITEMAP, which we use here inverted to return the status as a
        meaningful string. The five possible states are ( ordered from 0 to 4 ):
        RUNNING, REBOOTING, TERMINATED, PENDING & UNKNOWN.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )

        :return: S_OK( status ) | S_ERROR
        """

        result = self.getVMNode(nodeID)
        if not result["OK"]:
            return result

        state = result["Value"].state
        if state not in STATE_MAP:
            return S_ERROR("State %s not in STATEMAP" % state)

        return S_OK(STATE_MAP[state])

    def getVMNetwork(self, networkNames=None):
        """Get a network object corresponding to the networkName

        :param str networkName: network name
        :return: S_OK|S_ERROR network object in case of S_OK
        """
        if not networkNames:
            nameList = []
        else:
            nameList = list(networkNames)
        resultList = []
        if not nameList:
            nameList = self.parameters.get("networks")
            if not nameList:
                return S_ERROR("Network names are not specified")
            else:
                nameList = nameList.split(",")

        result = self.__driver.ex_list_networks()
        for oNetwork in result:
            if oNetwork.name in nameList:
                resultList.append(oNetwork)

        return S_OK(resultList)

    def stopVM(self, nodeID, publicIP=""):
        """
        Given the node ID it gets the node details, which are used to destroy the
        node making use of the libcloud.openstack driver. If three is any public IP
        ( floating IP ) assigned, frees it as well.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )
          **public_ip** - `string`
            public IP assigned to the node if any

        :return: S_OK | S_ERROR
        """

        # Get Node object with node details
        result = self.getVMNode(nodeID)
        if not result["OK"]:
            return result
        node = result["Value"]
        if node is None:
            # Node does not exist
            return S_OK()

        nodeIP = node.public_ips[0] if node.public_ips else None
        if not publicIP and nodeIP is not None:
            publicIP = nodeIP

        # Delete floating IP if any
        if publicIP:
            result = self.deleteFloatingIP(publicIP, node)
            if not result["OK"]:
                self.log.error("Failed in deleteFloatingIP:", result["Message"])

        # Destroy the VM instance
        if node is not None:
            try:
                result = self.__driver.destroy_node(node)
                if not result:
                    return S_ERROR("Failed to destroy node: %s" % node.id)
            except Exception as errmsg:
                return S_ERROR(errmsg)

        return S_OK()

    def getVMPool(self, poolName):

        try:
            poolList = self.__driver.ex_list_floating_ip_pools()
            for pool in poolList:
                if pool.name == poolName:
                    return S_OK(pool)
        except Exception as errmsg:
            return S_ERROR(errmsg)

        return S_ERROR("IP Pool with the name %s not found" % poolName)

    def assignFloatingIP(self, node):
        """
        Given a node, assign a floating IP from the ipPool defined on the imageConfiguration
        on the CS.

        :Parameters:
          **node** - `libcloud.compute.base.Node`
            node object with the vm details

        :return: S_OK( public_ip ) | S_ERROR
        """

        ipPool = self.parameters.get("ipPool")

        if ipPool:
            result = self.getVMPool(ipPool)
            if not result["OK"]:
                return result

            pool = result["Value"]
            try:
                floatingIP = pool.create_floating_ip()
                # Add sleep between creation and assignment
                time.sleep(60)
                self.__driver.ex_attach_floating_ip_to_node(node, floatingIP)
                publicIP = floatingIP.ip_address
                return S_OK(publicIP)

            except Exception as errmsg:
                return S_ERROR(errmsg)
        else:
            return S_ERROR("No IP pool specified")

    def getVMFloatingIP(self, publicIP):

        # We are still with IPv4
        publicIP = publicIP.replace("::ffff:", "")

        ipPool = self.parameters.get("ipPool")

        if ipPool:
            try:
                floatingIP = None
                poolList = self.__driver.ex_list_floating_ip_pools()
                for pool in poolList:
                    if pool.name == ipPool:
                        ipList = pool.list_floating_ips()
                        for ip in ipList:
                            if ip.ip_address == publicIP:
                                floatingIP = ip
                                break
                        break
                return S_OK(floatingIP)
            except Exception as errmsg:
                return S_ERROR(errmsg)
        else:
            return S_ERROR("No IP pool specified")

    def deleteFloatingIP(self, publicIP, node):
        """
        Deletes a floating IP <public_ip> from the server.

        :param str publicIP: public IP to be deleted
        :param object node: node to which IP is attached
        :return: S_OK | S_ERROR
        """

        # We are still with IPv4
        publicIP = publicIP.replace("::ffff:", "")

        result = self.getVMFloatingIP(publicIP)
        if not result["OK"]:
            return result

        floatingIP = result["Value"]
        if floatingIP is None:
            return S_OK()

        try:
            if node is not None:
                self.__driver.ex_detach_floating_ip_from_node(node, floatingIP)
            floatingIP.delete()
            return S_OK()
        except Exception as errmsg:
            return S_ERROR(errmsg)



#! /usr/bin/env python
########################################################################
# File :    dirac-stager-show-stats
# Author :  Daniela Remenska
########################################################################
"""
Reports breakdown of file(s) number/size in different staging states across Storage Elements.
Currently used Cache per SE is also reported. (active pins)

Example:
  $ dirac-stager-show-stats

   Status               SE                   NumberOfFiles        Size(GB)
  --------------------------------------------------------------------------
   Staged               GRIDKA-RDST          1                    4.5535
   StageSubmitted       GRIDKA-RDST          5                    22.586
   Waiting              PIC-RDST             3                    13.6478

  WARNING: the Size for files with Status=New is not yet determined at the point of selection!

   --------------------- current status of the SE Caches from the DB-----------
   GRIDKA-RDST    :      6 replicas with a size of 29.141 GB.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import gConfig, gLogger, exit as DIRACExit, S_OK, version


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=False)
    from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient

    client = StorageManagerClient()

    res = client.getCacheReplicasSummary()
    if not res["OK"]:
        gLogger.fatal(res["Message"])
        DIRACExit(2)
    stagerInfo = res["Value"]
    outStr = "\n"
    outStr += "  %s" % ("Status".ljust(20))
    outStr += "  %s" % ("SE".ljust(20))
    outStr += "  %s" % ("NumberOfFiles".ljust(20))
    outStr += "  %s" % ("Size(GB)".ljust(20))
    outStr += " \n--------------------------------------------------------------------------\n"
    if stagerInfo:
        for info in stagerInfo.values():
            outStr += "  %s" % (info["Status"].ljust(20))
            outStr += "  %s" % (info["SE"].ljust(20))
            outStr += "  %s" % (str(info["NumFiles"]).ljust(20))
            outStr += "  %s\n" % (str(info["SumFiles"]).ljust(20))
    else:
        outStr += "  %s" % ("Nothing to see here...Bye")
    outStr += "  \nWARNING: the Size for files with Status=New is not yet determined at the point of selection!\n"
    outStr += "--------------------- current status of the SE Caches from the DB-----------"
    res = client.getSubmittedStagePins()
    if not res["OK"]:
        gLogger.fatal(res["Message"])
        DIRACExit(2)
    storageElementUsage = res["Value"]
    if storageElementUsage:
        for storageElement in storageElementUsage.keys():
            seDict = storageElementUsage[storageElement]
            seDict["TotalSize"] = int(seDict["TotalSize"] / (1000 * 1000 * 1000.0))
            outStr += " \n %s: %s replicas with a size of %.3f GB." % (
                storageElement.ljust(15),
                str(seDict["Replicas"]).rjust(6),
                seDict["TotalSize"],
            )
    else:
        outStr += "  %s" % "\nStageRequest.getStorageUsage: No active stage/pin requests found."
    gLogger.notice(outStr)
    DIRACExit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-dms-lfn-accessURL
# Author :  Stuart Paterson
########################################################################
"""
Retrieve an access URL for an LFN replica given a valid DIRAC SE.

Example:
  $ dirac-dms-lfn-accessURL /formation/user/v/vhamar/Example.txt DIRAC-USER
  {'Failed': {},
   'Successful': {'/formation/user/v/vhamar/Example.txt': 'dips://dirac.in2p3.fr:9148/DataManagement/StorageElement\
   /formation/user/v/vhamar/Example.txt'}}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("LFN:      Logical File Name or file containing LFNs")
    Script.registerArgument("SE:       Valid DIRAC SE")
    Script.registerArgument("PROTO:    Optional protocol for accessURL", default=False, mandatory=False)
    _, args = Script.parseCommandLine(ignoreErrors=True)
    lfn, seName, proto = Script.getPositionalArgs(group=True)

    # pylint: disable=wrong-import-position
    from DIRAC.Interfaces.API.Dirac import Dirac

    if len(args) > 3:
        print("Only one LFN SE pair will be considered")

    dirac = Dirac()
    exitCode = 0

    try:
        with open(lfn, "r") as f:
            lfns = f.read().splitlines()
    except IOError:
        lfns = [lfn]

    for lfn in lfns:
        result = dirac.getAccessURL(lfn, seName, protocol=proto, printOutput=True)
        if not result["OK"]:
            print("ERROR: ", result["Message"])
            exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



"""Tests for the DIRAC.Core.Utilities.Extensions module"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
import six

import DIRAC
from DIRAC.Core.Utilities.Extensions import (
    findSystems,
    findAgents,
    findExecutors,
    findServices,
    findDatabases,
    extensionsByPriority,
    getExtensionMetadata,
)


def test_findSystems():
    systems = findSystems([DIRAC])
    assert len(systems) > 5
    assert all(system.endswith("System") for system in systems)


def test_findAgents():
    agents = findAgents([DIRAC])
    assert len(agents) > 5


def test_findExecutors():
    executors = findExecutors([DIRAC])
    assert len(executors) > 1


def test_findServices():
    services = findServices([DIRAC])
    assert len(services) > 5


def test_findDatabases():
    databases = findDatabases([DIRAC])
    assert len(databases) > 5
    assert all(str(fn).endswith(".sql") for system, fn in databases)


def test_extensionsByPriority():
    assert "DIRAC" in extensionsByPriority()


@pytest.mark.skipif(six.PY2, reason="Requires Python3")
def test_getExtensionMetadata():
    metadata = getExtensionMetadata("DIRAC")
    assert metadata["priority"] == 0



"""
This is a DIRAC WMS administrator interface.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import S_OK, S_ERROR

from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient


class WMSAdministratorHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, svcInfoDict):
        """WMS AdministratorService initialization"""
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobDB", "JobDB")
            if not result["OK"]:
                return result
            cls.jobDB = result["Value"]()
        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        cls.elasticJobParametersDB = None
        useESForJobParametersFlag = Operations().getValue("/Services/JobMonitoring/useESForJobParametersFlag", False)
        if useESForJobParametersFlag:
            try:
                result = ObjectLoader().loadObject(
                    "WorkloadManagementSystem.DB.ElasticJobParametersDB", "ElasticJobParametersDB"
                )
                if not result["OK"]:
                    return result
                cls.elasticJobParametersDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)

        cls.pilotManager = PilotManagerClient()

        return S_OK()

    types_setSiteMask = [list]

    def export_setSiteMask(self, siteList):
        """Set the site mask for matching. The mask is given in a form of Classad string.

        :param list siteList: site, status

        :return: S_OK()/S_ERROR()
        """
        credDict = self.getRemoteCredentials()
        maskList = [(site, "Active") for site in siteList]
        return self.jobDB.setSiteMask(maskList, credDict["DN"], "No comment")

    ##############################################################################
    types_getSiteMask = []

    @classmethod
    def export_getSiteMask(cls, siteState="Active"):
        """Get the site mask

        :param str siteState: site status

        :return: S_OK(list)/S_ERROR()
        """
        return cls.jobDB.getSiteMask(siteState)

    types_getSiteMaskStatus = []

    @classmethod
    def export_getSiteMaskStatus(cls, sites=None):
        """Get the site mask of given site(s) with columns 'site' and 'status' only

        :param sites: list of sites or site
        :type sites: list or str

        :return: S_OK()/S_ERROR() -- S_OK contain dict or str
        """
        return cls.jobDB.getSiteMaskStatus(sites)

    ##############################################################################
    types_getAllSiteMaskStatus = []

    @classmethod
    def export_getAllSiteMaskStatus(cls):
        """Get all the site parameters in the site mask

        :return: dict
        """
        return cls.jobDB.getAllSiteMaskStatus()

    ##############################################################################
    types_banSite = [six.string_types]

    def export_banSite(self, site, comment="No comment"):
        """Ban the given site in the site mask

        :param str site: site
        :param str comment: comment

        :return: S_OK()/S_ERROR()
        """
        credDict = self.getRemoteCredentials()
        author = credDict["username"] if credDict["username"] != "anonymous" else credDict["DN"]
        return self.jobDB.banSiteInMask(site, author, comment)

    ##############################################################################
    types_allowSite = [six.string_types]

    def export_allowSite(self, site, comment="No comment"):
        """Allow the given site in the site mask

        :param str site: site
        :param str comment: comment

        :return: S_OK()/S_ERROR()
        """
        credDict = self.getRemoteCredentials()
        author = credDict["username"] if credDict["username"] != "anonymous" else credDict["DN"]
        return self.jobDB.allowSiteInMask(site, author, comment)

    ##############################################################################
    types_clearMask = []

    @classmethod
    def export_clearMask(cls):
        """Clear up the entire site mask

        :return: S_OK()/S_ERROR()
        """
        return cls.jobDB.removeSiteFromMask(None)

    ##############################################################################
    types_getSiteMaskLogging = [six.string_types + (list,)]

    @classmethod
    def export_getSiteMaskLogging(cls, sites):
        """Get the site mask logging history

        :param list sites: sites

        :return: S_OK(dict)/S_ERROR()
        """
        if isinstance(sites, six.string_types):
            sites = [sites]

        return cls.jobDB.getSiteMaskLogging(sites)

    ##############################################################################
    types_getSiteMaskSummary = []

    @classmethod
    def export_getSiteMaskSummary(cls):
        """Get the mask status for all the configured sites

        :return: S_OK(dict)/S_ERROR()
        """
        # Get all the configured site names
        res = getSites()
        if not res["OK"]:
            return res
        sites = res["Value"]

        # Get the current mask status
        result = cls.jobDB.getSiteMaskStatus()
        siteDict = result["Value"]
        for site in sites:
            if site not in siteDict:
                siteDict[site] = "Unknown"

        return S_OK(siteDict)

    ##############################################################################
    types_getJobPilotOutput = [six.string_types + six.integer_types]

    def export_getJobPilotOutput(self, jobID):
        """Get the pilot job standard output and standard error files for the DIRAC
        job reference

        :param str jobID: job ID

        :return: S_OK(dict)/S_ERROR()
        """
        pilotReference = ""
        # Get the pilot grid reference first from the job parameters

        if self.elasticJobParametersDB:
            res = self.elasticJobParametersDB.getJobParameters(int(jobID), "Pilot_Reference")
            if not res["OK"]:
                return res
            if res["Value"].get(int(jobID)):
                pilotReference = res["Value"][int(jobID)]["Pilot_Reference"]

        if not pilotReference:
            res = self.jobDB.getJobParameter(int(jobID), "Pilot_Reference")
            if not res["OK"]:
                return res
            pilotReference = res["Value"]

        if not pilotReference:
            # Failed to get the pilot reference, try to look in the attic parameters
            res = self.jobDB.getAtticJobParameters(int(jobID), ["Pilot_Reference"])
            if res["OK"]:
                c = -1
                # Get the pilot reference for the last rescheduling cycle
                for cycle in res["Value"]:
                    if cycle > c:
                        pilotReference = res["Value"][cycle]["Pilot_Reference"]
                        c = cycle

        if pilotReference:
            return self.pilotManager.getPilotOutput(pilotReference)
        return S_ERROR("No pilot job reference found")

    ##############################################################################
    types_getSiteSummaryWeb = [dict, list, six.integer_types, six.integer_types]

    @classmethod
    def export_getSiteSummaryWeb(cls, selectDict, sortList, startItem, maxItems):
        """Get the summary of the jobs running on sites in a generic format

        :param dict selectDict: selectors
        :param list sortList: sorting list
        :param int startItem: start item number
        :param int maxItems: maximum of items

        :return: S_OK(dict)/S_ERROR()
        """
        return cls.jobDB.getSiteSummaryWeb(selectDict, sortList, startItem, maxItems)

    ##############################################################################
    types_getSiteSummarySelectors = []

    @classmethod
    def export_getSiteSummarySelectors(cls):
        """Get all the distinct selector values for the site summary web portal page

        :return: S_OK(dict)/S_ERROR()
        """
        resultDict = {}
        statusList = ["Good", "Fair", "Poor", "Bad", "Idle"]
        resultDict["Status"] = statusList
        maskStatus = ["Active", "Banned", "NoMask", "Reduced"]
        resultDict["MaskStatus"] = maskStatus

        res = getSites()
        if not res["OK"]:
            return res
        siteList = res["Value"]

        countryList = []
        for site in siteList:
            if site.find(".") != -1:
                country = site.split(".")[2].lower()
                if country not in countryList:
                    countryList.append(country)
        countryList.sort()
        resultDict["Country"] = countryList
        siteList.sort()
        resultDict["Site"] = siteList

        return S_OK(resultDict)
